@@ -1,4 +1,4 @@
 from . import act, base, gr00t, pi0
 from .act import ACT
-from .gr00t import Gr00tN1, Gr00tTrainerConfig, Gr00tTrainer
+from .gr00t import Gr00tN1, Gr00tTrainer, Gr00tTrainerConfig
 from .pi0 import Pi0
