@@ -1,15 +1,14 @@
 import asyncio
-from collections import deque
 import time
-from typing import Dict, List, Literal, Optional
+from collections import deque
+from typing import Any, Dict, List, Literal, Optional
 
-import cv2
 import httpx
 import json_numpy  # type: ignore
 import numpy as np
-from loguru import logger
 from fastapi import HTTPException
 from huggingface_hub import HfApi
+from loguru import logger
 from pydantic import BaseModel, Field, field_validator, model_validator
 
 from phosphobot.am.base import ActionModel
@@ -25,7 +24,6 @@ class InputFeature(BaseModel):
     shape: List[int]
 
 
-# Define the InputFeatures model to parse input_features
 class InputFeatures(BaseModel):
     state_key: str
     env_key: Optional[str] = None
@@ -40,7 +38,7 @@ class InputFeatures(BaseModel):
         return self.features[self.state_key].shape[0] // 6
 
     @model_validator(mode="before")
-    def infer_keys(cls, values):
+    def infer_keys(cls, values: Dict[str, Any]) -> Dict[str, Any]:
         """
         Preprocess input to infer state_key and video_keys if input is a flat features dict.
         Runs before field validation.
@@ -66,7 +64,7 @@ class InputFeatures(BaseModel):
         return values
 
     @field_validator("features", mode="before")
-    def validate_features(cls, value):
+    def validate_features(cls, value: Dict[str, Any]) -> Dict[str, InputFeature]:
         """
         Validate and transform the features dictionary into InputFeature instances.
         """
@@ -89,7 +87,7 @@ class InputFeatures(BaseModel):
         return result
 
     @model_validator(mode="after")
-    def validate_keys(self):
+    def validate_keys(self) -> "InputFeatures":
         """
         Validate state_key and video_keys against features after all fields are processed.
         """
@@ -177,8 +175,8 @@ class ACT(ActionModel):
         self,
         server_url: str = "http://localhost",
         server_port: int = 8080,
-        **kwargs,
-    ):
+        **kwargs: Any,
+    ) -> None:
         super().__init__(server_url, server_port)
         self.async_client = httpx.AsyncClient(
             base_url=server_url + f":{server_port}",
@@ -323,7 +321,7 @@ class ACT(ActionModel):
         model_id: str,
         all_cameras: AllCameras,
         robots: list[BaseManipulator],
-        cameras_keys_mapping: Dict[str, int] | None = None,
+        cameras_keys_mapping: Optional[Dict[str, int]] = None,
         verify_cameras: bool = True,
     ) -> ACTSpawnConfig:
         """
@@ -389,6 +387,8 @@ class ACT(ActionModel):
     def fetch_frame(
         cls, all_cameras: AllCameras, camera_id: int, resolution: list[int]
     ) -> np.ndarray:
+        import cv2
+
         rgb_frame = all_cameras.get_rgb_frame(
             camera_id=camera_id,
             resize=(resolution[2], resolution[1]),
@@ -420,14 +420,14 @@ class ACT(ActionModel):
         all_cameras: AllCameras,
         fps: int = 30,
         speed: float = 1.0,
-        cameras_keys_mapping: Dict[str, int] | None = None,
+        cameras_keys_mapping: Optional[Dict[str, int]] = None,
         prompt: Optional[str] = None,
         selected_camera_id: Optional[int] = None,
         angle_format: Literal["degrees", "radians", "other"] = "radians",
-        min_angle: float | None = None,
-        max_angle: float | None = None,
-        **kwargs,
-    ):
+        min_angle: Optional[float] = None,
+        max_angle: Optional[float] = None,
+        **kwargs: Any,
+    ) -> None:
         """
         AI control loop that runs in the background and sends actions to the robot.
         It uses the model to get the actions based on the current state of the robot and the cameras.
