@@ -2,7 +2,7 @@ import random
 import string
 from abc import ABC, abstractmethod
 from pathlib import Path
-from typing import Literal, Optional
+from typing import Any, Literal, Optional
 
 import av
 import numpy as np
@@ -12,6 +12,7 @@ from loguru import logger
 from pydantic import BaseModel, Field, field_validator, model_validator
 
 from phosphobot.models import InfoModel, ModelConfigurationResponse
+from phosphobot.utils import get_hf_token
 
 # Disable PyAV logs
 av.logging.set_level(None)
@@ -62,10 +63,10 @@ class ActionModel(ABC):
         Returns:
             np.ndarray: Sequence of actions (shape: [max_seq_length, n_actions]).
         """
-        raise NotImplementedError("""You cannot directly call the ActionModel class. 
+        raise NotImplementedError("""You cannot directly call the ActionModel class.
                                   You need to use an implementation ( ACT, PI0,...) or implement you own class.""")
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args: Any, **kwargs: Any) -> np.ndarray:
         """
         Makes the model instance callable, delegating to the forward method.
 
@@ -81,31 +82,31 @@ class ActionModel(ABC):
 
 class TrainingParamsAct(BaseModel):
     """
-    Training paramters are left to None by default and are set depending on the dataset in the training pipeline.
+    Training parameters are left to None by default and are set depending on the dataset in the training pipeline.
     """
 
-    batch_size: int | None = Field(
+    class Config:
+        extra = "allow"
+
+    batch_size: Optional[int] = Field(
         default=None,
-        description="Batch size for training, we run this on an A10G, leave it to None to auto-detect based on your dataset",
+        description="Batch size for training, we run this on an A10G. Leave it to None to auto-detect based on your dataset",
         gt=0,
         le=150,
     )
-    steps: int | None = Field(
+    steps: Optional[int] = Field(
         default=None,
-        description="Number of training steps, leave it to None to auto-detect based on your dataset",
+        description="Number of training steps. Leave it to None to auto-detect based on your dataset",
         gt=0,
         le=1_000_000,
     )
-    save_steps: int = Field(
+    save_freq: int = Field(
         default=5_000,
-        description="Number of steps between saving the model, leave it to None to get the default value",
+        description="Number of steps between saving the model.",
         gt=0,
         le=1_000_000,
     )
 
-    class Config:
-        extra = "forbid"
-
 
 DEFAULT_INSTRUCTION = "e.g.: green lego brick, red ball, blue plushy..."
 
@@ -151,7 +152,10 @@ class TrainingParamsActWithBbox(TrainingParamsAct):
         return instruction
 
 
-class TrainingParamsGr00T(BaseModel):
+class TrainingParamsPi0(BaseModel):
+    """
+    Training parameters for Pi0 model
+    """
     train_test_split: float = Field(
         default=1.0,
         description="Train test split ratio, default is 1.0 (no split), should be between 0 and 1",
@@ -165,7 +169,7 @@ class TrainingParamsGr00T(BaseModel):
 
     batch_size: int | None = Field(
         default=None,
-        description="Batch size for training, default is 64, decrease it if you get an out of memory error",
+        description="Batch size for training, leave it to None to auto-detect based on your dataset",
         gt=0,
         le=128,
     )
@@ -173,13 +177,7 @@ class TrainingParamsGr00T(BaseModel):
         default=10,
         description="Number of epochs to train for, default is 10",
         gt=0,
-        le=100,
-    )
-    save_steps: int = Field(
-        default=1_000,
-        description="Number of steps between saving the model, default is 1000",
-        gt=0,
-        le=100_000,
+        le=50,
     )
     learning_rate: float = Field(
         default=0.0001,
@@ -187,33 +185,75 @@ class TrainingParamsGr00T(BaseModel):
         gt=0,
         le=1,
     )
-
     data_dir: str = Field(
         default="data/", description="The directory to save the dataset to"
     )
-
-    validation_data_dir: str | None = Field(
-        default=None,
-        description="Optional directory to save the validation dataset to. If None, no validation will be done.",
-    )
-
     output_dir: str = Field(
         default="outputs/", description="The directory to save the model to"
     )
 
-    path_to_gr00t_repo: str = Field(
+    path_to_pi0_repo: str = Field(
         default=".",
-        description="The path to the Isaac-GR00T repo. If not provided, will assume we are in the repo.",
+        description="The path to the openpi repo. If not provided, will assume we are in the repo.",
     )
 
     class Config:
         extra = "forbid"
 
 
+class TrainingParamsGr00T(BaseModel):
+    class Config:
+        extra = "allow"
+
+    validation_dataset_name: Optional[str] = Field(
+        default=None,
+        description="Optional dataset repository ID on Hugging Face to use for validation",
+    )
+
+    batch_size: Optional[int] = Field(
+        default=64,
+        description="Batch size for training. Decrease it if you get an Out Of Memory (OOM) error",
+        gt=0,
+        le=128,
+        serialization_alias="batch-size",
+    )
+    num_epochs: int = Field(
+        default=10,
+        description="Number of epochs to train for.",
+        gt=0,
+        le=100,
+        serialization_alias="num-epochs",
+    )
+    save_steps: int = Field(
+        default=1_000,
+        description="Number of steps between saving the model.",
+        gt=0,
+        le=100_000,
+        serialization_alias="save-steps",
+    )
+    learning_rate: float = Field(
+        default=0.0001,
+        description="Learning rate for training.",
+        gt=0,
+        le=1,
+    )
+
+    data_dir: str = Field(
+        default="data/", description="The directory to save the dataset to"
+    )
+    validation_data_dir: Optional[str] = Field(
+        default=None,
+        description="Optional directory to save the validation dataset to. If None, validation is not run.",
+    )
+    output_dir: str = Field(
+        default="outputs/", description="The directory to save the model to"
+    )
+
+
 class BaseTrainerConfig(BaseModel):
-    model_type: Literal["ACT", "ACT_BBOX", "gr00t", "custom"] = Field(
+    model_type: Literal["ACT", "ACT_BBOX", "gr00t", "pi0", "custom"] = Field(
         ...,
-        description="Type of model to train, either 'ACT' or 'gr00t'",
+        description="Type of model to train, supports 'ACT', 'gr00t', and 'pi0'",
     )
     dataset_name: str = Field(
         ...,
@@ -229,120 +269,157 @@ class BaseTrainerConfig(BaseModel):
         description="WandB API key for tracking training, you can find it at https://wandb.ai/authorize",
     )
     training_params: Optional[
-        TrainingParamsAct | TrainingParamsActWithBbox | TrainingParamsGr00T
+        TrainingParamsAct | TrainingParamsActWithBbox | TrainingParamsGr00T | TrainingParamsPi0
     ] = Field(
         default=None,
-        description="Training parameters for the model, if not provided, default parameters will be used",
+        description="Training parameters for the model.",
     )
 
 
 class TrainingRequest(BaseTrainerConfig):
-    """Pydantic model for training request validation"""
+    """
+    Pydantic model for training request validation.
+    This version consolidates all model name and parameter logic into a single
+    validator to prevent redundant operations and fix the duplicate suffix bug.
+    """
+
+    private_mode: bool = Field(
+        default=False,
+        description="Whether to use private training (PRO users only)",
+    )
+    user_hf_token: Optional[str] = Field(
+        default=None,
+        description="User's personal HF token for private training",
+    )
 
     @model_validator(mode="before")
     @classmethod
-    def validate_required_fields(cls, data: dict) -> dict:
-        if not data.get("model_name"):
-            raise ValueError("model_name is required for training requests")
-        return data
-
-    @field_validator("model_name", mode="before")
-    def validate_model_name(cls, model_name: str) -> str:
-        if model_name is None:
-            raise ValueError("model_name is required for training requests")
-
-        # We add random characters to the model name to avoid collisions
-        random_chars = "".join(
-            random.choices(string.ascii_lowercase + string.digits, k=10)
-        )
-        # We need to make sure that the model is called phospho-app/...
-        # So we can upload it to the phospho Hugging Face repo
-        size = model_name.split("/")
-
-        def clamp_length(name: str, max_length: int) -> str:
-            """Clamp the length of the model name to a maximum length."""
-            if len(name) > max_length:
-                return name[:max_length]
-            return name
-
-        if len(size) == 1:
-            # Make sure the total model_name is shorter than 96 characters
-            model_name = clamp_length(
-                model_name, 96 - len(random_chars) - len("phospho-app/")
-            )
-
-            model_name = "phospho-app/" + model_name + "-" + random_chars
-        elif len(size) == 2:
-            if size[0] != "phospho-app":
-                # Make sure the total model_name is shorter than 96 characters
-                size[1] = clamp_length(
-                    size[1], 96 - len(random_chars) - len("phospho-app/")
-                )
-
-                model_name = "phospho-app/" + size[1] + "-" + random_chars
-        else:
-            raise ValueError(
-                "Model name should be in the format phospho-app/<model_name> or <model_name>",
-            )
-
-        return model_name
-
-    @field_validator("dataset_name", mode="before")
-    def validate_dataset(cls, dataset_name: str) -> str:
-        try:
-            url = f"https://huggingface.co/api/datasets/{dataset_name}/tree/main"
-            response = requests.get(url, timeout=5)
-            if response.status_code != 200:
-                raise ValueError()
-            return dataset_name
-        except Exception:
-            raise ValueError(
-                f"Dataset {dataset_name} is not a valid, public Hugging Face dataset. Please check the URL and try again. Your dataset name should be in the format <username>/<dataset_name>",
-            )
-
-    @model_validator(mode="before")
-    @classmethod
-    def validate_training_params(cls, data: dict) -> dict:
+    def prepare_model_and_params(cls, data: dict) -> dict:
+        """
+        Consolidated validator that handles all pre-validation logic:
+        1. Validates `model_type` and initializes `training_params`.
+        2. Determines the correct namespace (public or private).
+        3. Implements the new logic for handling or generating the model name.
+        """
+        # Step 1: Validate model_type and training_params
         model_type_to_class: dict[str, type[BaseModel]] = {
             "ACT": TrainingParamsAct,
             "ACT_BBOX": TrainingParamsActWithBbox,
             "gr00t": TrainingParamsGr00T,
+            "pi0": TrainingParamsPi0,
         }
-
         model_type = data.get("model_type")
-        training_params = data.get("training_params")
-
-        if model_type is None:
-            raise ValueError(
-                "Model type is required. Please provide a valid model type: 'ACT', 'ACT_BBOX' or 'gr00t'."
-            )
+        if not model_type:
+            raise ValueError("Model type is required.")
         if model_type not in model_type_to_class:
-            raise ValueError(
-                f"Unsupported model type: {model_type}. Valid options are: {list(model_type_to_class.keys())}"
-            )
+            raise ValueError(f"Unsupported model type: {model_type}.")
 
         params_class = model_type_to_class[model_type]
-
-        if training_params:
-            logger.debug(
-                f"Training parameters provided: {training_params}, validating them with {params_class.__name__}"
+        if "training_params" in data and data["training_params"]:
+            data["training_params"] = params_class.model_validate(
+                data["training_params"]
             )
-            data["training_params"] = params_class.model_validate(training_params)
+        else:
+            data["training_params"] = params_class()  # Set defaults
 
-        if training_params is None:
-            # If no training params are provided, we set the default ones
-            data["training_params"] = params_class()
+        # Step 2: Determine the namespace and validate user token
+        namespace = "phospho-app"
+        api = None
+        if data.get("private_mode"):
+            token = data.get("user_hf_token") or get_hf_token()
+            if not token:
+                raise ValueError("Private training requires a valid HF token.")
+            data["user_hf_token"] = token
+            try:
+                api = HfApi(token=token)
+                user_info = api.whoami()
+                username = user_info.get("name")
+                if not username:
+                    raise ValueError("Could not get username from HF token.")
+                namespace = username
+            except Exception as e:
+                raise ValueError(f"Failed to validate user namespace: {e}")
+
+        # Step 3 & 4: Handle model name based on the new business rules
+        user_provided_name = data.get("model_name")
+        max_len = 96  # Max length for a model name
+
+        if user_provided_name:
+            # Rule 1: If a model name is provided, use it.
+            # The base name is the user-provided name.
+            base_name = user_provided_name
+
+            # If the namespace is different from phospho-app and not in private mode, raise error
+            if namespace != "phospho-app" and not data.get("private_mode"):
+                raise ValueError(
+                    "You can only use a custom model name in private mode or with the phospho-app namespace."
+                )
+
+            # Ensure the name is less than the max length
+            if len(base_name) > max_len:
+                base_name = base_name[:max_len]
+
+            data["model_name"] = base_name
+
+        else:
+            # If no model name is provided, generate one.
+            dataset_name = data.get("dataset_name")
+            if not dataset_name or len(dataset_name.split("/")) != 2:
+                raise ValueError(
+                    "A valid dataset_name is required to generate a model name."
+                )
+            dataset_base_name = dataset_name.split("/")[1]
+            base_name = f"{model_type}-{dataset_base_name}"
+
+            random_chars = "".join(
+                random.choices(string.ascii_lowercase + string.digits, k=10)
+            )
+
+            # Clamp the base_name length to ensure the final name is valid
+            max_base_len = (
+                max_len - len(namespace) - len(random_chars) - 2
+            )  # for "/" and "-"
+            clamped_base_name = base_name[:max_base_len]
+
+            data["model_name"] = f"{namespace}/{clamped_base_name}-{random_chars}"
 
         return data
 
+    @model_validator(mode="after")
+    def validate_dataset_access(self) -> "TrainingRequest":
+        """Validate dataset access based on private mode (no changes needed here)."""
+        if self.private_mode:
+            if not self.user_hf_token:
+                raise ValueError("Private training requires a user HF token.")
+            try:
+                api = HfApi(token=self.user_hf_token)
+                api.repo_info(repo_id=self.dataset_name, repo_type="dataset")
+            except Exception as e:
+                raise ValueError(f"Cannot access dataset {self.dataset_name}: {e}")
+        else:
+            try:
+                url = (
+                    f"https://huggingface.co/api/datasets/{self.dataset_name}/tree/main"
+                )
+                response = requests.get(url, timeout=5)
+                response.raise_for_status()
+            except Exception:
+                raise ValueError(
+                    f"Dataset {self.dataset_name} is not a valid, public dataset."
+                )
+
+        return self
+
 
 class HuggingFaceTokenValidator:
     @staticmethod
-    def has_write_access(hf_token: str, hf_model_name: str) -> bool:
+    def has_write_access(hf_token: str, hf_model_name: str, private: bool) -> bool:
         """Check if the HF token has write access by attempting to create a repo."""
-        api = HfApi()
+        api = HfApi(token=hf_token)
         try:
-            api.create_repo(hf_model_name, private=False, exist_ok=True, token=hf_token)
+            api.create_repo(
+                hf_model_name, private=private, exist_ok=True, token=hf_token
+            )
             return True  # The token has write access
         except Exception as e:
             print(f"Write access check failed: {e}")
@@ -352,14 +429,12 @@ class HuggingFaceTokenValidator:
 def generate_readme(
     model_type: str,
     dataset_repo_id: str,
-    folder_path: Path | None = None,
+    training_params: BaseModel,
+    folder_path: Optional[Path] = None,
     wandb_run_url: Optional[str] = None,
-    steps: Optional[int] = None,
-    epochs: Optional[int] = None,
-    batch_size: Optional[int] = None,
     error_traceback: Optional[str] = None,
     return_readme_as_bytes: bool = False,
-):
+) -> Path | bytes:
     readme = f"""
 ---
 datasets: {dataset_repo_id}
@@ -370,10 +445,13 @@ tags:
 - phosphobot
 - {model_type}
 task_categories:
-- robotics                                               
+- robotics
 ---
 
-# {model_type} Model - phospho Training Pipeline
+# {model_type} model - 🧪 phosphobot training pipeline
+
+- **Dataset**: [{dataset_repo_id}](https://huggingface.co/datasets/{dataset_repo_id})
+- **Wandb run id**: {wandb_run_url}
 
 """
     if error_traceback:
@@ -388,20 +466,18 @@ We faced an issue while training your model.
 """
     else:
         readme += """
-## This model was trained using **phospho**.
+## This model was trained using **[🧪phospho](https://phospho.ai)**
 
 Training was successful, try it out on your robot!
 
 """
 
     readme += f"""
-## Training parameters:
+## Training parameters
 
-- **Dataset**: [{dataset_repo_id}](https://huggingface.co/datasets/{dataset_repo_id})
-- **Wandb run URL**: {wandb_run_url}
-- **Epochs**: {epochs}
-- **Batch size**: {batch_size}
-- **Training steps**: {steps}
+```text
+{training_params.model_dump_json(indent=2)}
+```
 
 📖 **Get Started**: [docs.phospho.ai](https://docs.phospho.ai?utm_source=huggingface_readme)
 
@@ -423,7 +499,7 @@ Training was successful, try it out on your robot!
 def resize_dataset(
     dataset_root_path: Path,
     resize_to: tuple = (320, 240),
-) -> tuple[bool, bool, str | None]:
+) -> tuple[bool, bool, Optional[str]]:
     """
     Resize the dataset to a smaller size for faster training.
 
@@ -528,9 +604,9 @@ def resize_dataset(
 
 class BaseTrainer(ABC):
     """
-    Currently only implemented for gr00t.
+    Abstract class for training models.
     """
 
     @abstractmethod
-    def train(self):
+    def train(self) -> None:
         pass
