@@ -8,12 +8,11 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass
 from io import BytesIO
 from pathlib import Path
-from typing import Any, Callable, Dict, Literal, Tuple
+from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union
 
-import cv2
-import zmq
 import numpy as np
 import pandas as pd
+import zmq
 from fastapi import HTTPException
 from huggingface_hub import HfApi, snapshot_download
 from loguru import logger
@@ -69,7 +68,7 @@ class BaseInferenceServer:
     Can add custom endpoints by calling `register_endpoint`.
     """
 
-    def __init__(self, host: str = "*", port: int = 5555):
+    def __init__(self, host: str = "*", port: int = 5555) -> None:
         self.running = True
         self.context = zmq.Context()
         self.socket = self.context.socket(zmq.REP)
@@ -80,7 +79,7 @@ class BaseInferenceServer:
         self.register_endpoint("ping", self._handle_ping, requires_input=False)
         self.register_endpoint("kill", self._kill_server, requires_input=False)
 
-    def _kill_server(self):
+    def _kill_server(self) -> None:
         """
         Kill the server.
         """
@@ -94,7 +93,7 @@ class BaseInferenceServer:
 
     def register_endpoint(
         self, name: str, handler: Callable, requires_input: bool = True
-    ):
+    ) -> None:
         """
         Register a new endpoint to the server.
 
@@ -186,7 +185,7 @@ class RobotInferenceServer(BaseInferenceServer):
     Server with three endpoints for real robot policies
     """
 
-    def __init__(self, model: BasePolicy, host: str = "*", port: int = 5555):
+    def __init__(self, model: BasePolicy, host: str = "*", port: int = 5555) -> None:
         super().__init__(host, port)
         self.register_endpoint("get_action", model.get_action)
         self.register_endpoint(
@@ -194,7 +193,7 @@ class RobotInferenceServer(BaseInferenceServer):
         )
 
     @staticmethod
-    def start_server(policy: BasePolicy, port: int):
+    def start_server(policy: BasePolicy, port: int) -> None:
         server = RobotInferenceServer(policy, port=port)
         server.run()
 
@@ -202,7 +201,7 @@ class RobotInferenceServer(BaseInferenceServer):
 class BaseInferenceClient:
     def __init__(
         self, host: str = "localhost", port: int = 5555, timeout_ms: int = 15000
-    ):
+    ) -> None:
         self.context = zmq.Context()
 
         self.host = host
@@ -211,7 +210,7 @@ class BaseInferenceClient:
         self.version = 2
         self._init_socket()
 
-    def _init_socket(self):
+    def _init_socket(self) -> None:
         """Initialize or reinitialize the socket with current settings"""
         self.socket = self.context.socket(zmq.REQ)
         self.socket.connect(f"tcp://{self.host}:{self.port}")
@@ -224,14 +223,14 @@ class BaseInferenceClient:
             self._init_socket()  # Recreate socket for next attempt
             return False
 
-    def kill_server(self):
+    def kill_server(self) -> None:
         """
         Kill the server.
         """
         self.call_endpoint("kill", requires_input=False)
 
     def call_endpoint(
-        self, endpoint: str, data: dict | None = None, requires_input: bool = True
+        self, endpoint: str, data: Optional[Dict] = None, requires_input: bool = True
     ) -> dict:
         """
         Call an endpoint on the server.
@@ -264,7 +263,7 @@ class BaseInferenceClient:
             # legacy: the handler's own dict
             return resp
 
-    def __del__(self):
+    def __del__(self) -> None:
         """Cleanup resources on destruction"""
         self.socket.close()
         self.context.term()
@@ -303,7 +302,7 @@ class ComponentStatistics(BaseModel):
         extra = "allow"
 
     @model_validator(mode="before")
-    def collect_active_components(self):
+    def collect_active_components(self) -> "ComponentStatistics":
         """
         Collect names and values of all fields containing valid Stats before validation.
         Ensures extra fields are included in active_components.
@@ -407,7 +406,7 @@ class HuggingFaceModelConfig(BaseModel):
     # This will store the found embodiment config
     embodiment: EmbodimentConfig
     # This will store the original field name
-    embodiment_field_name: str | None = None
+    embodiment_field_name: Optional[str] = None
 
     class Config:
         extra = "allow"
@@ -488,8 +487,8 @@ class Gr00tN1(ActionModel):
         ],  # These values are read from the values in experiment_cfg/metadata.json
         server_url: str = "localhost",
         server_port: int = 5555,
-        **kwargs,
-    ):
+        **kwargs: Any,
+    ) -> None:
         super().__init__(server_url, server_port)
         self.client = ExternalRobotInferenceClient(host=server_url, port=server_port)
         self.action_keys = action_keys
@@ -638,7 +637,7 @@ class Gr00tN1(ActionModel):
         model_id: str,
         all_cameras: AllCameras,
         robots: list[BaseManipulator],
-        cameras_keys_mapping: Dict[str, int] | None = None,
+        cameras_keys_mapping: Optional[Dict[str, int]] = None,
         verify_cameras: bool = True,
     ) -> Gr00tSpawnConfig:
         """
@@ -704,18 +703,18 @@ class Gr00tN1(ActionModel):
     async def control_loop(
         self,
         control_signal: AIControlSignal,
-        robots: list[BaseManipulator],
+        robots: List[BaseManipulator],
         model_spawn_config: Gr00tSpawnConfig,
         all_cameras: AllCameras,
-        prompt: str | None = None,
+        prompt: Optional[str] = None,
         fps: int = 30,
         speed: float = 1.0,
-        cameras_keys_mapping: Dict[str, int] | None = None,
+        cameras_keys_mapping: Optional[Dict[str, int]] = None,
         unit: Literal["degrees", "rad", "other"] = "rad",
-        min_angle: float | None = None,
-        max_angle: float | None = None,
+        min_angle: Optional[float] = None,
+        max_angle: Optional[float] = None,
         **kwargs: Any,
-    ):
+    ) -> None:
         """
         AI control loop that runs in the background and sends actions to the robot.
         It uses the model to get the actions based on the current state of the robot and the cameras.
@@ -723,6 +722,8 @@ class Gr00tN1(ActionModel):
         The loop runs at the specified fps and speed.
         """
 
+        import cv2
+
         nb_iter = 0
         config = model_spawn_config.hf_model_config
         signal_marked_as_started = False
@@ -933,29 +934,28 @@ class Gr00tTrainerConfig(BaseTrainerConfig):
     training_params: TrainingParamsGr00T
 
 
-def check_for_nans_null_in_value(value):
+def check_for_nans_null_in_value(value: Union[list, tuple, pd.DataFrame]) -> bool:
     """
     Check if a value contains NaN/null, including nested lists
     """
-    if pd.isna(value).any():
-        return True
-
-    if pd.isnull(value).any():
-        return True
-
     if isinstance(value, (list, tuple)):
         for item in value:
             if check_for_nans_null_in_value(item):
                 return True
+    else:
+        if pd.isna(value).any():
+            return True
+        if pd.isnull(value).any():
+            return True
 
     return False
 
 
-def check_parquet_files(folder_path):
+def check_parquet_files(folder_path: Path) -> None:
     """
     Check all parquet files in a folder for NaN/null values in the action/observation column
 
-    Will raise an error if it finds any NaN/null values in the action/observation column.
+    Raise an error if any NaN/null values in the action/observation column.
     """
     folder_path = Path(folder_path)
 
@@ -1024,7 +1024,7 @@ def check_parquet_files(folder_path):
         )
 
 
-def generate_modality_json(data_dir) -> tuple[int, int]:
+def generate_modality_json(data_dir: Path) -> Tuple[int, int]:
     # Load the metadata file to get image keys
     with open(data_dir / "meta" / "info.json", "r") as f:
         metadata = json.load(f)
@@ -1074,106 +1074,6 @@ def generate_modality_json(data_dir) -> tuple[int, int]:
     return number_of_robots, number_of_cameras
 
 
-async def run_gr00t_training(
-    data_dir,
-    output_dir,
-    batch_size,
-    epochs,
-    number_of_robots,
-    number_of_cameras,
-    learning_rate,
-    save_steps: int,
-    wandb_enabled: bool,
-    validation_data_dir=None,
-    timeout_seconds: int | None = None,
-    gr00t_repo_path: str = ".",
-):
-    cmd = [
-        "python",
-        f"{gr00t_repo_path}/scripts/gr00t_finetune.py",
-        "--dataset-path",
-        str(data_dir),
-    ]
-
-    if validation_data_dir is not None:
-        logger.info(f"Using validation dataset from {validation_data_dir}")
-        cmd.extend(["--validation-dataset-path", str(validation_data_dir)])
-    else:
-        logger.info("No validation dataset provided. No validation will be done.")
-
-    # Add remaining arguments
-    cmd.extend(
-        [
-            # Only 1 GPU for now
-            # Open an issue for multi-GPU support
-            "--num-gpus",
-            "1",
-            "--output-dir",
-            str(output_dir),
-            "--batch-size",
-            str(batch_size),
-            "--num-epochs",
-            str(epochs),
-            "--save-steps",
-            str(save_steps),
-            "--num-arms",
-            str(number_of_robots),
-            "--num-cams",
-            str(number_of_cameras),
-            "--learning_rate",
-            str(learning_rate),
-            "--report_to",
-            "wandb" if wandb_enabled else "tensorboard",
-            "--video_backend",
-            "torchvision_av",
-        ]
-    )
-
-    logger.info(f"Starting training with command: {' '.join(cmd)}")
-
-    process = await asyncio.create_subprocess_exec(
-        *cmd,
-        stdout=asyncio.subprocess.PIPE,
-        stderr=asyncio.subprocess.STDOUT,
-        # 512 KB buffer size, default is 64 but seems to be too small
-        limit=512 * 1024,
-    )
-
-    output_lines = []
-
-    async def read_output():
-        assert process.stdout is not None
-        async for line in process.stdout:
-            stripped_line = line.decode().strip()
-            print(stripped_line)
-            output_lines.append(stripped_line)
-
-    try:
-        if timeout_seconds is None:
-            # No timeout
-            await read_output()
-        else:
-            # Timeout
-            await asyncio.wait_for(read_output(), timeout=timeout_seconds)
-    except asyncio.TimeoutError:
-        process.kill()
-        await process.wait()
-        logger.error(f"Training process timed out after {timeout_seconds} seconds.")
-        raise TimeoutError(
-            f"Training process exceeded timeout of {timeout_seconds} seconds. Please consider lowering the number of epochs and/or batch size."
-        )
-
-    await process.wait()
-
-    if process.returncode != 0:
-        error_output = "\n".join(output_lines[-10:])
-        error_msg = f"Training process failed with exit code {process.returncode}:\n{error_output}"
-        logger.error(error_msg)
-        raise RuntimeError(error_msg)
-
-    return output_lines
-
-
 class Gr00tTrainer(BaseTrainer):
     """
     Trainer for the Gr00t model.
@@ -1183,7 +1083,12 @@ class Gr00tTrainer(BaseTrainer):
     def __init__(self, config: Gr00tTrainerConfig):
         self.config = config
 
-    def train(self, timeout_seconds: int | None = None) -> None:
+    def train(
+        self,
+        timeout_seconds: Optional[int] = None,
+        private_mode: bool = False,
+        hf_token: Optional[str] = None,
+    ) -> None:
         """
         You can pass a timeout in seconds to the training process.
         If the training process exceeds this time, it will be
@@ -1202,13 +1107,15 @@ class Gr00tTrainer(BaseTrainer):
 
         if self.config.model_name is not None:
             # We check if the user has write access to the model-id
-            hf_token = os.getenv("HF_TOKEN")
+            hf_token = hf_token or os.getenv("HF_TOKEN")
             if hf_token is None:
                 raise ValueError(
                     "HF_TOKEN environment variable is not set. Please set it to your Hugging Face token."
                 )
             if not HuggingFaceTokenValidator().has_write_access(
-                hf_token=hf_token, hf_model_name=self.config.model_name
+                hf_token=hf_token,
+                hf_model_name=self.config.model_name,
+                private=private_mode,
             ):
                 raise ValueError(
                     f"The provided HF token does not have write access to {self.config.model_name}"
@@ -1224,7 +1131,7 @@ class Gr00tTrainer(BaseTrainer):
                     repo_type="dataset",
                     revision=selected_branch,
                     local_dir=str(data_dir),
-                    token=os.getenv("HF_TOKEN"),
+                    token=hf_token,
                 )
                 DATASET_PATH = Path(dataset_path_as_str)
                 logger.info(
@@ -1256,7 +1163,7 @@ class Gr00tTrainer(BaseTrainer):
         logger.info("Generating modality.json file")
         number_of_robots, number_of_cameras = generate_modality_json(data_dir)
 
-        val_data_dir: Path | None = None
+        val_data_dir: Optional[Path] = None
         if self.config.training_params.validation_dataset_name is not None:
             if self.config.training_params.validation_data_dir is not None:
                 val_data_dir = Path(self.config.training_params.validation_data_dir)
@@ -1271,8 +1178,8 @@ class Gr00tTrainer(BaseTrainer):
                         repo_id=self.config.training_params.validation_dataset_name,
                         repo_type="dataset",
                         revision=selected_branch,
-                        local_dir=str(val_data_dir),
-                        token=os.getenv("HF_TOKEN"),
+                        local_dir=(val_data_dir),
+                        token=hf_token,
                     )
                     VAL_DATASET_PATH = Path(dataset_path_val_str)
                     logger.info(
@@ -1306,36 +1213,14 @@ class Gr00tTrainer(BaseTrainer):
             # We set the validation data dir to None to avoid passing it to the training script
             val_data_dir = None
 
-        # Find the total number of frames in the dataset in meta / info.json
-        with open(data_dir / "meta" / "info.json", "r") as f:
-            info = json.load(f)
-            total_frames = info["total_frames"]
-
-        steps = (
-            total_frames
-            * self.config.training_params.epochs
-            // self.config.training_params.batch_size
-            + 1
-        )
-
-        logger.info(
-            f"Will train for {self.config.training_params.epochs} epochs, which is {steps} steps"
-        )
-
         asyncio.run(
-            run_gr00t_training(
+            self._call_training_script(
                 data_dir=data_dir,
                 output_dir=output_dir,
-                batch_size=self.config.training_params.batch_size,
-                epochs=self.config.training_params.epochs,
+                validation_data_dir=val_data_dir,
                 number_of_robots=number_of_robots,
                 number_of_cameras=number_of_cameras,
-                learning_rate=self.config.training_params.learning_rate,
-                save_steps=self.config.training_params.save_steps,
-                wandb_enabled=self.config.wandb_api_key is not None,
-                validation_data_dir=val_data_dir,
                 timeout_seconds=timeout_seconds,
-                gr00t_repo_path=self.config.training_params.path_to_gr00t_repo,
             )
         )
         logger.info("Training finished")
@@ -1344,13 +1229,13 @@ class Gr00tTrainer(BaseTrainer):
             logger.info(f"Uploading model to Hugging Face as {self.config.model_name}")
 
             # Upload using huggingface_hub
-            api = HfApi(token=os.getenv("HF_TOKEN"))
+            api = HfApi(token=hf_token)
             files_directory = output_dir
 
             api.create_repo(
                 repo_id=self.config.model_name,
                 repo_type="model",
-                private=False,
+                private=private_mode,
                 exist_ok=True,
             )
 
@@ -1420,8 +1305,7 @@ class Gr00tTrainer(BaseTrainer):
             readme = generate_readme(
                 model_type="gr00t",
                 dataset_repo_id=self.config.dataset_name,
-                epochs=self.config.training_params.epochs,
-                batch_size=self.config.training_params.batch_size,
+                training_params=self.config.training_params,
                 return_readme_as_bytes=True,
             )
 
@@ -1435,12 +1319,110 @@ class Gr00tTrainer(BaseTrainer):
             # Get the model URL
             huggingface_model_url = f"https://huggingface.co/{self.config.model_name}"
             logger.info(f"Model successfully uploaded to {huggingface_model_url}")
-
-            logger.info("Model successfully uploaded to Hugging Face")
+        else:
             logger.info(
-                f"Find your trained model on Hugging Face: {huggingface_model_url}"
+                "Skipping upload to Hugging Face. Provide a model-id to enable automatic upload to Hugging Face"
             )
 
+    async def _call_training_script(
+        self,
+        data_dir: Path,
+        output_dir: Path,
+        validation_data_dir: Optional[Path],
+        number_of_robots: int,
+        number_of_cameras: int,
+        timeout_seconds: Optional[int] = None,
+        gr00t_repo_path: str = ".",
+    ) -> List[str]:
+        training_params = self.config.training_params
+        wandb_enabled = self.config.wandb_api_key is not None
+
+        cmd = [
+            "python",
+            f"{gr00t_repo_path}/scripts/gr00t_finetune.py",
+            "--dataset-path",
+            str(data_dir),
+        ]
+
+        if validation_data_dir is not None:
+            logger.info(f"Using validation dataset from {validation_data_dir}")
+            cmd.extend(["--validation-dataset-path", str(validation_data_dir)])
         else:
-            logger.info("Skipping upload to Hugging Face")
-            logger.info("Provide a model-id to enable automatic upload to Hugging Face")
+            logger.info("No validation dataset provided. No validation will be done.")
+
+        # Add remaining arguments
+        cmd.extend(
+            [
+                # Only 1 GPU for now
+                # Open an issue for multi-GPU support
+                "--num-gpus",
+                "1",
+                "--output-dir",
+                str(output_dir),
+                "--num-arms",
+                str(number_of_robots),
+                "--num-cams",
+                str(number_of_cameras),
+                "--report_to",
+                "wandb" if wandb_enabled else "tensorboard",
+                "--video_backend",
+                "torchvision_av",
+            ]
+        )
+
+        # Adds all extra parameters from the training_params
+        training_params_dict = training_params.model_dump(
+            by_alias=True,
+            exclude_none=True,
+            exclude={
+                "data_dir": True,
+                "output_dir": True,
+                "validation_data_dir": True,
+            },
+        )
+        for key, value in training_params_dict.items():
+            cmd.extend([f"--{key}", str(value)])
+
+        logger.info(f"Starting training with command: {' '.join(cmd)}")
+
+        process = await asyncio.create_subprocess_exec(
+            *cmd,
+            stdout=asyncio.subprocess.PIPE,
+            stderr=asyncio.subprocess.STDOUT,
+            # 512 KB buffer size, default is 64 but seems to be too small
+            limit=512 * 1024,
+        )
+
+        output_lines = []
+
+        async def read_output() -> None:
+            assert process.stdout is not None
+            async for line in process.stdout:
+                stripped_line = line.decode().strip()
+                print(stripped_line)
+                output_lines.append(stripped_line)
+
+        try:
+            if timeout_seconds is None:
+                # No timeout
+                await read_output()
+            else:
+                # Timeout
+                await asyncio.wait_for(read_output(), timeout=timeout_seconds)
+        except asyncio.TimeoutError:
+            process.kill()
+            await process.wait()
+            logger.error(f"Training process timed out after {timeout_seconds} seconds.")
+            raise TimeoutError(
+                f"Training process exceeded timeout of {timeout_seconds} seconds. Please consider lowering the number of epochs and/or batch size."
+            )
+
+        await process.wait()
+
+        if process.returncode != 0:
+            error_output = "\n".join(output_lines[-10:])
+            error_msg = f"Training process failed with exit code {process.returncode}:\n{error_output}"
+            logger.error(error_msg)
+            raise RuntimeError(error_msg)
+
+        return output_lines
