@@ -1,5 +1,8 @@
 import asyncio
 import atexit
+import base64
+import binascii
+import json
 import platform
 import subprocess
 import threading
@@ -7,7 +10,17 @@ import time
 from abc import ABC, abstractmethod
 from functools import lru_cache
 from pathlib import Path
-from typing import AsyncGenerator, Dict, List, Literal, Optional, Tuple, cast
+from typing import (
+    Any,
+    AsyncGenerator,
+    Dict,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    Tuple,
+    cast,
+)
 
 import cv2
 import numpy as np
@@ -186,8 +199,8 @@ def detect_camera_type(
 
 # TODO: Handle multiple realsense cameras
 def _find_cameras(
-    possible_camera_ids: list[int],
-    raise_when_empty=False,
+    possible_camera_ids: List[int],
+    raise_when_empty: bool = False,
     camera_names: List[str] = [],
 ) -> list[int]:
     """
@@ -198,8 +211,6 @@ def _find_cameras(
     This ignores realsense cameras.
     """
 
-    import cv2
-
     camera_ids = []
     for camera_idx in possible_camera_ids:
         try:
@@ -246,7 +257,7 @@ def _find_cameras(
 
 # TODO: Handle multiple realsense cameras
 def detect_video_indexes(
-    max_index_search_range: int | None = None,
+    max_index_search_range: Optional[int] = None,
     mock: bool = False,
     camera_names: List[str] = [],
 ) -> list[int]:
@@ -301,10 +312,10 @@ class BaseCamera(ABC):
     height: int
     fps: int
 
-    def __init__(self):
+    def __init__(self) -> None:
         atexit.register(self.stop)
 
-    def __del__(self):
+    def __del__(self) -> None:
         self.stop()
 
     @property
@@ -313,7 +324,7 @@ class BaseCamera(ABC):
 
     @abstractmethod
     def get_rgb_frame(
-        self, resize: tuple[int, int] | None = None
+        self, resize: Optional[tuple[int, int]] = None
     ) -> Optional[cv2.typing.MatLike]:
         """Get the latest frame from the camera."""
         pass
@@ -329,10 +340,10 @@ class BaseCamera(ABC):
 
     def get_jpeg_rgb_frame(
         self,
-        target_size: tuple[int, int] | None,
-        quality: int | None,
+        target_size: Optional[tuple[int, int]],
+        quality: Optional[int],
         is_video_frame: bool = True,
-    ) -> bytes | None:
+    ) -> Optional[bytes]:
         if is_video_frame:
             rgb_frame = self.get_rgb_frame(resize=target_size)
         else:
@@ -352,10 +363,10 @@ class BaseCamera(ABC):
 
     async def generate_rgb_frames(
         self,
-        target_size: tuple[int, int] | None,
-        quality: int | None,
+        target_size: Optional[tuple[int, int]],
+        quality: Optional[int],
         is_video_frame: bool = True,
-        request: Request | None = None,
+        request: Optional[Request] = None,
     ) -> AsyncGenerator:
         """Generator for video frames"""
         try:
@@ -391,9 +402,6 @@ class BaseCamera(ABC):
             logger.warning(f"{self.camera_name} Error generating frames: {str(e)}")
             self.stop()
 
-    async def generate_depth_frames(self):
-        return await self.generate_rgb_frames(is_video_frame=False)
-
 
 class VideoCamera(threading.Thread, BaseCamera):
     camera_type: CameraTypes = "classic"
@@ -529,7 +537,7 @@ Camera type: {self.camera_type}""")
                     self.last_frame = frame
 
     def get_rgb_frame(
-        self, resize: tuple[int, int] | None = None
+        self, resize: Optional[tuple[int, int]] = None
     ) -> Optional[cv2.typing.MatLike]:
         """
         Read a frame from the camera. Returns None if the frame could not be read.
@@ -542,7 +550,7 @@ Camera type: {self.camera_type}""")
         if self.last_frame is None:
             logger.warning(f"{self.camera_name}: No frame available")
 
-        frame: np.ndarray | None = None
+        frame: Optional[np.ndarray] = None
         # Convert from BGR to RGB
         if self.last_frame is not None:
             frame = cv2.cvtColor(self.last_frame, cv2.COLOR_BGR2RGB)
@@ -566,13 +574,13 @@ class DummyCamera(VideoCamera):
         width: int = 640,
         height: int = 480,
         fps: int = 30,
-    ):
+    ) -> None:
         self.width = width
         self.height = height
         self.fps = fps
         super().__init__(camera_type=camera_type)
 
-    def init_camera(self):
+    def init_camera(self) -> bool:
         """
         The simulated camera cannot be opened with opencv, so we return True.
         """
@@ -580,7 +588,7 @@ class DummyCamera(VideoCamera):
         return True
 
     def get_rgb_frame(
-        self, resize: tuple[int, int] | None = None
+        self, resize: Optional[Tuple[int, int]] = None
     ) -> Optional[cv2.typing.MatLike]:
         """
         Read a frame from the camera. Returns None if the frame could not be read.
@@ -623,7 +631,7 @@ class StereoCamera(VideoCamera):
         return f"StereoCamera {self.camera_id}"
 
     def get_left_eye_rgb_frame(
-        self, resize: tuple[int, int] | None = None
+        self, resize: Optional[Tuple[int, int]] = None
     ) -> Optional[cv2.typing.MatLike]:
         last_frame = self.get_rgb_frame()
         if last_frame is None:
@@ -636,7 +644,7 @@ class StereoCamera(VideoCamera):
         return left_frame
 
     def get_right_eye_rgb_frame(
-        self, resize: tuple[int, int] | None = None
+        self, resize: Optional[Tuple[int, int]] = None
     ) -> Optional[cv2.typing.MatLike]:
         last_frame = self.get_rgb_frame()
         if last_frame is None:
@@ -737,13 +745,12 @@ try:
                 config.enable_device(self.device_serial)
 
                 # Configure streams
+                # Intialize at 30FPS
                 config.enable_stream(
-                    stream_type=rs.stream.color,
-                    format=rs.format.bgr8,
+                    stream_type=rs.stream.color, format=rs.format.bgr8, framerate=30
                 )
                 config.enable_stream(
-                    stream_type=rs.stream.depth,
-                    format=rs.format.z16,
+                    stream_type=rs.stream.depth, format=rs.format.z16, framerate=30
                 )
 
                 # Add a small delay to ensure device is ready
@@ -792,7 +799,7 @@ try:
             return f"RealsenseCamera {self.device_index} ({self.device_serial})"
 
         def get_rgb_frame(
-            self, resize: tuple[int, int] | None = None
+            self, resize: Optional[Tuple[int, int]] = None
         ) -> Optional[cv2.typing.MatLike]:
             # To get the video frame, get the couple (video, depth) frame from the wait_for_frames method
             if not self.is_active:
@@ -813,7 +820,7 @@ try:
             return frame
 
         def get_depth_frame(
-            self, resize: tuple[int, int] | None = None
+            self, resize: Optional[Tuple[int, int]] = None
         ) -> Optional[cv2.typing.MatLike]:
             # To get the depth frame, also get the couple (video, depth) frame from the wait_for_frames method
             # The method get_depth and get_rgb_frame can be called simultaneously without lagging (tested)
@@ -868,11 +875,11 @@ try:
             return self.realsense_camera.is_active
 
         @is_active.setter
-        def is_active(self, value):
+        def is_active(self, value: bool) -> None:
             return
 
         def get_rgb_frame(
-            self, resize: Optional[tuple[int, int]] = None
+            self, resize: Optional[Tuple[int, int]] = None
         ) -> Optional[cv2.typing.MatLike]:
             if not self.is_active:
                 return None
@@ -908,22 +915,23 @@ except ImportError:
     REALSENSE_AVAILABLE = False
 
     class RealSenseCamera(BaseCamera):  # type: ignore
-        def __init__(self, *args, **kwargs):
+        def __init__(self, *args: Iterable[Any], **kwargs: Dict[str, Any]) -> None:
             raise ImportError("Install pyrealsense2 to add RealSense camera support.")
 
     class RealSenseVirtualCamera(VideoCamera):  # type: ignore
-        def __init__(self, *args, **kwargs):
+        def __init__(self, *args: Iterable[Any], **kwargs: Dict[str, Any]) -> None:
             raise ImportError("Install pyrealsense2 to add RealSense camera support.")
 
 
 class ZMQCamera(VideoCamera):
     """
-    A camera class that subscribes to a ZeroMQ PUB socket to receive video frames.
-    It dynamically detects stream properties (width, height) from the first frame received.
+    A camera that connects to a ZMQ PUSH socket, receiving JSON-serialized
+    data with base64-encoded camera frames and performing manual topic filtering.
     """
 
     camera_type: CameraTypes = "zmq"
     connect_to: str
+    topic: Optional[str]
     stream_initialized: bool = False
     context: Optional[zmq.Context] = None
     socket: Optional[zmq.Socket] = None
@@ -932,46 +940,49 @@ class ZMQCamera(VideoCamera):
     def __init__(
         self,
         connect_to: str = "tcp://localhost:5555",
+        topic: Optional[str] = None,
         disable: bool = False,
         camera_id: Optional[int] = None,
     ):
         self.connect_to = connect_to
+        self.topic = topic if topic and topic.strip() else None
         self.stream_initialized = False
-
-        # Now call the parent constructor. It will safely call our overridden init_camera.
         super().__init__(video=None, disable=disable, camera_id=camera_id)
+        self.last_frame = None
+        self.lock = threading.Lock()
+        self._stop_event = threading.Event()
+        self.thread = None
 
     @property
     def camera_name(self) -> str:
-        return f"ZMQCamera {self.connect_to}"
+        if self.topic:
+            return f"ZMQCamera(addr='{self.connect_to}', topic='{self.topic}')"
+        return f"ZMQCamera(addr='{self.connect_to}', receives_all_topics)"
 
     def init_camera(self) -> bool:
-        """
-        Overrides the parent method to initialize a ZMQ connection.
-        Sets width and height to 0, as they are unknown until the first frame arrives.
-        """
+        """Initializes the ZMQ PULL socket."""
         try:
-            logger.info(
-                f"{self.camera_name}: Connecting to ZMQ publisher at {self.connect_to}"
-            )
+            logger.info(f"{self.camera_name}: Connecting to ZMQ PUSH server...")
             self.context = zmq.Context()
-            self.socket = self.context.socket(zmq.SUB)
-            # Set a timeout for receiving messages to prevent blocking indefinitely
+            self.socket = self.context.socket(zmq.PULL)
             self.socket.setsockopt(zmq.RCVTIMEO, 2000)
             self.socket.connect(self.connect_to)
-            self.socket.setsockopt(zmq.SUBSCRIBE, b"")
+
+            if self.topic:
+                logger.info(
+                    f"{self.camera_name}: Will manually filter for topic '{self.topic}'."
+                )
+            else:
+                logger.warning(
+                    f"{self.camera_name}: No topic set, will process all received messages."
+                )
 
             self.poller = zmq.Poller()
             self.poller.register(self.socket, zmq.POLLIN)
+            self.width, self.height = 0, 0
+            self.fps = 30
 
-            # Initialize properties as unknown. They will be updated by the run() thread.
-            self.width = 0
-            self.height = 0
-            self.fps = 30  # Keep a default, or it could also be sent by the publisher
-
-            logger.success(
-                f"{self.camera_name}: ZMQ connection established. Waiting for first frame..."
-            )
+            logger.success(f"{self.camera_name}: ZMQ PULL connection established.")
             return True
         except Exception as e:
             logger.error(
@@ -979,69 +990,72 @@ class ZMQCamera(VideoCamera):
             )
             return False
 
+    def _process_frame_data(self, data: dict) -> None:
+        """Helper function to process a received data dictionary and update the frame."""
+        if not self.stream_initialized:
+            shape = data["shape"]
+            self.height, self.width, _ = shape
+            self.stream_initialized = True
+            logger.success(
+                f"{self.camera_name}: Stream properties detected: {self.width}x{self.height}"
+            )
+
+        frame_bytes = base64.b64decode(data["frame_bytes"])
+        frame = np.frombuffer(frame_bytes, dtype=np.dtype(data["dtype"]))
+        reconstructed_frame = frame.reshape(data["shape"])
+        with self.lock:
+            self.last_frame = cv2.cvtColor(reconstructed_frame, cv2.COLOR_RGB2BGR)
+
     def run(self) -> None:
-        """
-        Overrides the thread's run method.
-        This loop polls the ZMQ socket for new frames and initializes stream properties from the first frame.
-        """
-        if not self.is_active or not self.socket or not self.context or not self.poller:
+        """Polls the ZMQ PULL socket and manually filters messages by topic."""
+        if not self.socket or not self.poller:
+            logger.error(f"{self.camera_name}: Cannot run, ZMQ socket not initialized.")
             return
 
         while not self._stop_event.is_set():
-            # Poll the socket with a timeout to allow the stop event to be checked
             socks = dict(self.poller.poll(timeout=100))
-
             if self.socket in socks and socks[self.socket] == zmq.POLLIN:
                 try:
-                    data = self.socket.recv_pyobj(flags=zmq.NOBLOCK)
+                    message_parts = self.socket.recv_multipart(flags=zmq.NOBLOCK)
+                    if len(message_parts) < 2:
+                        continue  # Ignore malformed messages
 
-                    # --- DYNAMIC INITIALIZATION LOGIC ---
-                    if not self.stream_initialized:
-                        shape = data["shape"]
-                        # Update the camera's properties from the received metadata
-                        self.height, self.width, _ = shape
-                        self.stream_initialized = True
-                        logger.success(
-                            f"{self.camera_name}: Stream properties detected from first frame: {self.width}x{self.height}"
-                        )
-                    # --- END DYNAMIC INITIALIZATION LOGIC ---
+                    received_topic = message_parts[0].decode()
 
-                    # Reconstruct the numpy array from the received bytes and metadata
-                    frame = np.frombuffer(
-                        data["frame_bytes"], dtype=np.dtype(data["dtype"])
-                    )
-                    reconstructed_frame = frame.reshape(data["shape"])
-
-                    with self.lock:
-                        # The parent VideoCamera expects a BGR frame from OpenCV. Since our stream
-                        # is RGB, we must convert it before storing it in self.last_frame.
-                        self.last_frame = cv2.cvtColor(
-                            reconstructed_frame, cv2.COLOR_RGB2BGR
-                        )
+                    # Process if this message is for us, or if we accept all topics
+                    if self.topic is None or received_topic == self.topic:
+                        json_bytes = message_parts[1]
+                        data = json.loads(json_bytes.decode("utf-8"))
+                        self._process_frame_data(data)
 
                 except zmq.Again:
-                    # No message was ready to be received, continue polling
                     continue
-                except (KeyError, IndexError, TypeError) as e:
+                except (
+                    json.JSONDecodeError,
+                    binascii.Error,
+                    KeyError,
+                    IndexError,
+                    TypeError,
+                ) as e:
                     logger.warning(
-                        f"{self.camera_name}: Received malformed or unexpected data packet. Error: {e}"
+                        f"{self.camera_name}: Malformed data packet. Error: {e}"
                     )
                 except Exception as e:
-                    logger.warning(f"{self.camera_name}: Error receiving frame: {e}")
+                    logger.error(
+                        f"{self.camera_name}: Unexpected error processing frame: {e}"
+                    )
 
         logger.info(f"{self.camera_name}: Thread stopped.")
 
     def stop(self) -> None:
-        """
-        Extends the parent stop method to gracefully close ZMQ resources.
-        """
-        if not self.is_active:
+        """Extends the parent stop method to gracefully close ZMQ resources."""
+        if self._stop_event.is_set():
             return
-
         logger.debug(f"{self.camera_name}: Stopping...")
-        super().stop()  # This sets the _stop_event
+        self._stop_event.set()
+        if self.thread and self.thread.is_alive():
+            self.thread.join(timeout=2.0)
 
-        # Clean up ZMQ resources
         try:
             if self.socket:
                 self.socket.close()
@@ -1052,19 +1066,19 @@ class ZMQCamera(VideoCamera):
 
 
 class AllCameras:
-    disabled_cameras: list[int] | None
+    disabled_cameras: Optional[List[int]]
     video_cameras: List[VideoCamera]
     realsense_cameras: List[RealSenseCamera]
     zmq_cameras: List[ZMQCamera]
 
     camera_ids: List[int]
     camera_names: List[str]
-    _main_camera: BaseCamera | None = None
+    _main_camera: Optional[BaseCamera] = None
     # If it's None, record everything. Otherwise, record only the corresponding cameras
     _cameras_ids_to_record: List[int]
     _is_detecting: bool = False
 
-    def __init__(self, disabled_cameras: list[int] | None = None):
+    def __init__(self, disabled_cameras: Optional[List[int]] = None):
         """
         AllCameras class to manage all cameras connected to the computer.
         Args:
@@ -1106,7 +1120,7 @@ class AllCameras:
 
         return all_cameras
 
-    def detect_cameras(self):
+    def detect_cameras(self) -> None:
         """
         Detect all cameras connected to the computer and initialize them.
         """
@@ -1160,9 +1174,6 @@ class AllCameras:
                     video=cv2.VideoCapture(index),
                     disable=self.disabled_cameras is not None
                     and index in self.disabled_cameras,
-                    width=1280,
-                    height=480,
-                    fps=30,
                     camera_id=index,
                 )
                 # Set the camera_id to the first position and reindex
@@ -1234,7 +1245,7 @@ class AllCameras:
         self._cameras_ids_to_record = self.camera_ids
         self._is_detecting = False
 
-    def add_custom_camera(self, camera: BaseCamera):
+    def add_custom_camera(self, camera: BaseCamera) -> None:
         """
         Manually adds an initialized custom camera instance to the list of active cameras.
         This is useful for adding virtual or non-discoverable cameras like ZMQCamera.
@@ -1434,7 +1445,7 @@ class AllCameras:
             ],
         )
 
-    def stop(self):
+    def stop(self) -> None:
         for camera in self.cameras:
             camera.stop()
 
@@ -1453,7 +1464,7 @@ class AllCameras:
     def get_rgb_frame(
         self,
         camera_id: Optional[int] = None,
-        resize: Optional[tuple[int, int]] = None,
+        resize: Optional[Tuple[int, int]] = None,
     ) -> Optional[cv2.typing.MatLike]:
         """
         Return the latest RGB frame from the specidied camera
@@ -1479,26 +1490,27 @@ class AllCameras:
         return frame
 
     def get_rgb_frames_for_all_cameras(
-        self, resize: Optional[tuple[int, int]] = None
-    ) -> Dict[str, cv2.typing.MatLike | None]:
+        self, resize: Optional[Tuple[int, int]] = None
+    ) -> Dict[str, Optional[cv2.typing.MatLike]]:
         """
         Get the RGB frames for all the cameras.
 
         Returns a dict with the camera id as key and the frame as value.
         """
-        frames: dict[str, cv2.typing.MatLike | None] = {}
+        frames: Dict[str, Optional[cv2.typing.MatLike]] = {}
         for camera in self.video_cameras:
+            if camera.is_active is False:
+                continue
+
             if camera.camera_type == "stereo":
                 camera = cast(StereoCamera, camera)
                 left_side = camera.get_left_eye_rgb_frame(resize=resize)
                 right_side = camera.get_right_eye_rgb_frame(resize=resize)
-                frames[f"camera_{len(frames)}"] = left_side
-                frames[f"camera_{len(frames)}"] = right_side
-            elif camera.camera_type == "classic":
-                frame = camera.get_rgb_frame(resize=resize)
-                frames[f"camera_{len(frames)}"] = frame
+                frames[f"{camera.camera_id}_left"] = left_side
+                frames[f"{camera.camera_id}_right"] = right_side
             else:
-                logger.warning(f"Unknown camera type: {camera.camera_type}")
+                frame = camera.get_rgb_frame(resize=resize)
+                frames[f"{camera.camera_id}"] = frame
         return frames
 
     @property
@@ -1509,7 +1521,7 @@ class AllCameras:
         return self._cameras_ids_to_record
 
     @cameras_ids_to_record.setter
-    def cameras_ids_to_record(self, camera_ids: List[int] | None):
+    def cameras_ids_to_record(self, camera_ids: Optional[List[int]]) -> None:
         """
         Set the camera ids to record.
         """
@@ -1523,7 +1535,7 @@ class AllCameras:
             self._cameras_ids_to_record = self.camera_ids
 
     @property
-    def main_camera(self) -> BaseCamera | None:
+    def main_camera(self) -> Optional[BaseCamera]:
         """
         Get the main camera among the selected camera ids.
         If selected camera ids are not provided, we select amoung all available cameras.
@@ -1548,7 +1560,7 @@ class AllCameras:
         return self._main_camera
 
     @main_camera.setter
-    def main_camera(self, camera: BaseCamera):
+    def main_camera(self, camera: BaseCamera) -> None:
         """
         Explicitly set the main camera.
         """
@@ -1556,8 +1568,8 @@ class AllCameras:
 
     def get_main_camera_frames(
         self,
-        target_video_size: tuple[int, int],
-    ) -> List[cv2.typing.MatLike] | None:
+        target_video_size: Tuple[int, int],
+    ) -> Optional[List[cv2.typing.MatLike]]:
         """
         Get the frames from the main camera.
         """
@@ -1610,7 +1622,7 @@ class AllCameras:
 
     def get_secondary_camera_frames(
         self,
-        target_video_size: tuple[int, int],
+        target_video_size: Tuple[int, int],
     ) -> List[cv2.typing.MatLike]:
         """
         Get the frames from every camera except the main camera.
@@ -1699,7 +1711,7 @@ def get_all_cameras() -> AllCameras:
     return cameras
 
 
-def get_all_cameras_no_init() -> AllCameras | None:
+def get_all_cameras_no_init() -> Optional[AllCameras]:
     """
     Return the global AllCameras instance without initializing it.
     This is useful for testing purposes.
