@@ -2,11 +2,12 @@ import asyncio
 import json
 import traceback
 from copy import copy
-from typing import cast
+from typing import Optional, cast
 
 import httpx
 import json_numpy  # type: ignore
 import numpy as np
+import serial
 from fastapi import (
     APIRouter,
     BackgroundTasks,
@@ -17,12 +18,13 @@ from fastapi import (
 )
 from loguru import logger
 from scipy.spatial.transform import Rotation as R
+from supabase_auth.types import Session as SupabaseSession
 
 from phosphobot.ai_control import CustomAIControlSignal, setup_ai_control
 from phosphobot.camera import AllCameras, get_all_cameras
 from phosphobot.control_signal import ControlSignal
 from phosphobot.hardware.base import BaseManipulator
-from phosphobot.leader_follower import RobotPair, leader_follower_loop
+from phosphobot.leader_follower import RobotPair, start_leader_follower_loop
 from phosphobot.models import (
     AIControlStatusResponse,
     AIStatusResponse,
@@ -38,11 +40,14 @@ from phosphobot.models import (
     RelativeEndEffectorPosition,
     RobotConfigResponse,
     RobotConnectionRequest,
+    RobotConnectionResponse,
     SpawnStatusResponse,
     StartAIControlRequest,
     StartLeaderArmControlRequest,
     StartServerRequest,
     StatusResponse,
+    TeleopSettings,
+    TeleopSettingsRequest,
     TemperatureReadResponse,
     TemperatureWriteRequest,
     TorqueControlRequest,
@@ -51,7 +56,6 @@ from phosphobot.models import (
     VoltageReadResponse,
 )
 from phosphobot.robot import (
-    PiperHardware,
     RemotePhosphobot,
     RobotConnectionManager,
     SO100Hardware,
@@ -86,9 +90,9 @@ signal_vr_control = ControlSignal()
     description="Initialize the robot to its initial position before starting the teleoperation.",
 )
 async def move_init(
-    robot_id: int | None = None,
+    robot_id: Optional[int] = None,
     teleop_manager: TeleopManager = Depends(get_teleop_manager),
-):
+) -> StatusResponse:
     """
     Initialize the robot to its initial position before starting the teleoperation.
     """
@@ -104,7 +108,7 @@ async def move_init(
 )
 async def move_teleop_post(
     control_data: AppControlData,
-    robot_id: int | None = None,
+    robot_id: Optional[int] = None,
     teleop_manager: TeleopManager = Depends(get_teleop_manager),
 ) -> StatusResponse:
     teleop_manager.robot_id = robot_id
@@ -118,7 +122,7 @@ async def move_teleop_ws(
     websocket: WebSocket,
     rcm: RobotConnectionManager = Depends(get_rcm),
     teleop_manager: TeleopManager = Depends(get_teleop_manager),
-):
+) -> None:
     teleop_manager.robot_id = None
 
     if not await rcm.robots:
@@ -147,7 +151,7 @@ async def move_teleop_ws(
 async def move_teleop_udp(
     udp_server: UDPServer = Depends(get_udp_server),
     teleop_manager: TeleopManager = Depends(get_teleop_manager),
-):
+) -> UDPServerInformationResponse:
     """
     Start a UDP server to send and receive teleoperation data to the robot.
     """
@@ -159,7 +163,7 @@ async def move_teleop_udp(
 @router.post("/move/teleop/udp/stop", response_model=StatusResponse)
 async def stop_teleop_udp(
     udp_server: UDPServer = Depends(get_udp_server),
-):
+) -> StatusResponse:
     """
     Stop the UDP server main loop.
     """
@@ -223,11 +227,6 @@ async def move_to_absolute_position(
             if robot.name == "so100":
                 # We invert rx and ry
                 target_controller_orientation = np.array([query.ry, query.rx, query.rz])
-            elif robot.name == "agilex-piper":
-                rotation = R.from_euler("y", -90, degrees=True)
-                target_controller_orientation = rotation.apply(
-                    [query.rx, query.ry, query.rz]
-                )
             else:
                 target_controller_orientation = np.array([query.rx, query.ry, query.rz])
 
@@ -252,7 +251,7 @@ async def move_to_absolute_position(
         else:
             orientation_residual = 0
 
-        async def try_moving_to_target():
+        async def try_moving_to_target() -> None:
             nonlocal \
                 current_position, \
                 current_orientation, \
@@ -271,13 +270,12 @@ async def move_to_absolute_position(
                 await robot.move_robot_absolute(
                     target_position=target_position,
                     target_orientation_rad=target_orientation_rad,
-                    interpolate_trajectory=False,
                 )
                 current_position, current_orientation = robot.forward_kinematics()
                 position_residual = np.linalg.norm(current_position - target_position)
                 if use_angles:
-                    orientation_residual = np.linalg.norm(
-                        current_orientation - target_orientation_rad
+                    orientation_residual = float(
+                        np.linalg.norm(current_orientation - target_orientation_rad)
                     )
 
         await try_moving_to_target()
@@ -477,7 +475,7 @@ async def move_sleep(
     description="Retrieve the position, orientation, and open status of the robot's end effector. Only available for manipulators.",
 )
 async def end_effector_read(
-    query: EndEffectorReadRequest | None = None,
+    query: Optional[EndEffectorReadRequest] = None,
     robot_id: int = 0,
     rcm: RobotConnectionManager = Depends(get_rcm),
 ) -> EndEffectorPosition:
@@ -486,7 +484,7 @@ async def end_effector_read(
     """
     robot = await rcm.get_robot(robot_id)
     if query is None:
-        query = EndEffectorReadRequest(sync=False)
+        query = EndEffectorReadRequest(only_gripper=False, sync=False)
 
     if not isinstance(robot, BaseManipulator):
         raise HTTPException(
@@ -494,12 +492,28 @@ async def end_effector_read(
             detail="Robot is not a manipulator and does not have an end effector",
         )
 
+    if query.only_gripper:
+        if not hasattr(robot, "closing_gripper_value"):
+            raise HTTPException(
+                status_code=400,
+                detail=f"Robot {robot.name} does not have a .closing_gripper_value attribute to read the gripper state.",
+            )
+        return EndEffectorPosition(
+            x=None,
+            y=None,
+            z=None,
+            rx=None,
+            ry=None,
+            rz=None,
+            open=robot.closing_gripper_value,
+        )
+
     initial_position = getattr(robot, "initial_position", None)
     initial_orientation_rad = getattr(robot, "initial_orientation_rad", None)
     if initial_position is None or initial_orientation_rad is None:
         raise HTTPException(
             status_code=400,
-            detail=f"Call /move/init before using this endpoint for robot {robot.name}. ",
+            detail=f"Before using /end-effector/read you need to call /move/init?robot_id={robot_id} to initialize the robot's position and orientation.",
         )
 
     position, orientation, open_status = robot.get_end_effector_state(sync=query.sync)
@@ -530,7 +544,7 @@ async def end_effector_read(
 async def read_voltage(
     robot_id: int = 0,
     rcm: RobotConnectionManager = Depends(get_rcm),
-):
+) -> VoltageReadResponse:
     """
     Read voltage of the robot.
     """
@@ -556,7 +570,7 @@ async def read_voltage(
 async def read_temperature(
     robot_id: int = 0,
     rcm: RobotConnectionManager = Depends(get_rcm),
-):
+) -> TemperatureReadResponse:
     """
     Read temperature of the robot.
     """
@@ -609,7 +623,7 @@ async def write_temperature(
 async def read_torque(
     robot_id: int = 0,
     rcm: RobotConnectionManager = Depends(get_rcm),
-):
+) -> TorqueReadResponse:
     """
     Read torque of the robot.
     """
@@ -639,7 +653,7 @@ async def read_torque(
 )
 async def toggle_torque(
     request: TorqueControlRequest,
-    robot_id: int | None = None,
+    robot_id: Optional[int] = None,
     rcm: RobotConnectionManager = Depends(get_rcm),
 ) -> StatusResponse:
     """
@@ -684,7 +698,7 @@ async def toggle_torque(
     description="Read the current positions of the robot's joints in radians and motor units.",
 )
 async def read_joints(
-    request: JointsReadRequest | None = None,
+    request: Optional[JointsReadRequest] = None,
     robot_id: int = 0,
     rcm: RobotConnectionManager = Depends(get_rcm),
 ) -> JointsReadResponse:
@@ -815,7 +829,7 @@ async def start_leader_follower(
     if signal_leader_follower.is_in_loop():
         raise HTTPException(
             status_code=400,
-            detail="Leader-follower control is already running. Call /move/leader/stop first.",
+            detail="Leader-follower control is already running. To stop it, call /move/leader/stop",
         )
 
     # Parse the robot IDs from the request
@@ -847,11 +861,7 @@ async def start_leader_follower(
                     detail=f"Follower must be an instance of SO100Hardware for robot pair {i}.",
                 )
         else:
-            valid_robot_types = (
-                SO100Hardware,
-                PiperHardware,
-                RemotePhosphobot,
-            )
+            valid_robot_types = (BaseManipulator, RemotePhosphobot)
             if not isinstance(leader, valid_robot_types):
                 raise HTTPException(
                     status_code=400,
@@ -871,7 +881,7 @@ async def start_leader_follower(
 
     # Add background task to run the control loop
     background_tasks.add_task(
-        background_task_log_exceptions(leader_follower_loop),
+        start_leader_follower_loop,
         robot_pairs=robot_pairs,
         control_signal=signal_leader_follower,
         invert_controls=request.invert_controls,
@@ -978,7 +988,7 @@ async def spawn_inference_server(
     query: StartServerRequest,
     rcm: RobotConnectionManager = Depends(get_rcm),
     all_cameras: AllCameras = Depends(get_all_cameras),
-    session=Depends(user_is_logged_in),
+    session: SupabaseSession = Depends(user_is_logged_in),
 ) -> SpawnStatusResponse:
     """
     Start an inference server and return the server info.
@@ -1029,7 +1039,7 @@ async def start_ai_control(
     background_tasks: BackgroundTasks,
     rcm: RobotConnectionManager = Depends(get_rcm),
     all_cameras: AllCameras = Depends(get_all_cameras),
-    session=Depends(user_is_logged_in),
+    session: SupabaseSession = Depends(user_is_logged_in),
 ) -> AIControlStatusResponse:
     """
     Start the auto control by AI
@@ -1156,7 +1166,7 @@ async def start_ai_control(
 async def stop_ai_control(
     background_tasks: BackgroundTasks,
     rcm: RobotConnectionManager = Depends(get_rcm),
-    session=Depends(user_is_logged_in),
+    session: SupabaseSession = Depends(user_is_logged_in),
 ) -> StatusResponse:
     """
     Stop the auto control by AI
@@ -1166,7 +1176,7 @@ async def stop_ai_control(
 
     # Call the /stop endpoint in Modal
     @background_task_log_exceptions
-    async def stop_modal():
+    async def stop_modal() -> None:
         async with httpx.AsyncClient(timeout=5) as client:
             await client.post(
                 url=f"{tokens.MODAL_API_URL}/stop",
@@ -1227,15 +1237,12 @@ async def resume_ai_control(
 @router.post(
     "/ai-control/feedback",
     response_model=StatusResponse,
-    summary="Feedback about the auto control session",
+    summary="Feedback about the AI control session",
 )
 async def feedback_ai_control(
     request: FeedbackRequest,
-    session=Depends(user_is_logged_in),
+    session: SupabaseSession = Depends(user_is_logged_in),
 ) -> StatusResponse:
-    """
-    Feedback about the auto control session
-    """
     supabase_client = await get_client()
 
     await (
@@ -1248,27 +1255,79 @@ async def feedback_ai_control(
     return StatusResponse(message="Feedback sent")
 
 
-@router.post("/robot/add-connection", response_model=StatusResponse)
+@router.post("/robot/add-connection", response_model=RobotConnectionResponse)
 async def add_robot_connection(
     query: RobotConnectionRequest,
     rcm: RobotConnectionManager = Depends(get_rcm),
+) -> RobotConnectionResponse:
+    """
+    Manually add a robot connection to the robot manager.
+    Useful for adding robot that are accessible only via WiFi, for example.
+    """
+    try:
+        robot_id, robot = await rcm.add_connection(
+            robot_name=query.robot_name,
+            connection_details=query.connection_details,
+        )
+        return RobotConnectionResponse(
+            status="ok",
+            message=f"Robot connection to {query.robot_name} added",
+            robot_id=robot_id,
+        )
+    except serial.SerialException as e:
+        if "Access is denied" in str(e):
+            logger.warning(
+                f"Failed to add robot connection: {e}\n{traceback.format_exc()}"
+            )
+            raise HTTPException(
+                status_code=403,
+                detail=f"Permission error: {e}. If you're on Windows, try running with WSL (Windows Subsystem for Linux) or phosphobot has the authorization to use the serial port.",
+            )
+        elif "Permission denied" in str(e):
+            logger.warning(
+                f"Failed to add robot connection: {e}\n{traceback.format_exc()}"
+            )
+            raise HTTPException(
+                status_code=403,
+                detail=f"Permission denied: {e}. If you're on Linux, try running phosphobot as sudo (`sudo phosphobot`) and ensure that your user has access to the serial port.",
+            )
+        else:
+            logger.error(
+                f"Failed to add robot connection (SerialException): {e}\n{traceback.format_exc()}"
+            )
+            raise HTTPException(
+                status_code=400, detail=f"Failed to add robot connection: {e}"
+            )
+    except Exception as e:
+        logger.error(f"Failed to add robot connection: {e}\n{traceback.format_exc()}")
+        raise HTTPException(
+            status_code=400, detail=f"Failed to add robot connection: {e}"
+        )
+
+
+@router.post("/robot/disconnect", response_model=StatusResponse)
+async def disconnect_robot(
+    robot_id: int = 0,
+    rcm: RobotConnectionManager = Depends(get_rcm),
 ) -> StatusResponse:
     """
     Manually add a robot connection to the robot manager.
     Useful for adding robot that are accessible only via WiFi, for example.
     """
     try:
-        await rcm.add_connection(
-            robot_name=query.robot_name,
-            connection_details=query.connection_details,
-        )
+        await rcm.remove_connection(robot_id=robot_id)
         return StatusResponse(
-            status="ok", message=f"Robot connection to {query.robot_name} added"
+            status="ok",
+            message=f"Robot connection to {robot_id} removed successfully",
         )
+    except HTTPException as e:
+        raise e
     except Exception as e:
-        logger.error(f"Failed to add robot connection: {e}\n{traceback.format_exc()}")
+        logger.error(
+            f"Failed to remove robot connection {robot_id}: {e}\n{traceback.format_exc()}"
+        )
         raise HTTPException(
-            status_code=400, detail=f"Failed to add robot connection: {e}"
+            status_code=400, detail=f"Failed to remove robot connection {robot_id}: {e}"
         )
 
 
@@ -1297,3 +1356,44 @@ async def get_robot_config(
         status_code=400,
         detail=f"Robot {robot.name} does not support configuration retrieval.",
     )
+
+
+@router.post(
+    "/teleop/settings/read",
+    response_model=TeleopSettings,
+    summary="Read Teleop Settings",
+    description="Get current teleoperation settings.",
+)
+async def read_teleop_settings(
+    teleop_manager: TeleopManager = Depends(get_teleop_manager),
+) -> TeleopSettings:
+    """
+    Get current teleoperation settings.
+    """
+    return teleop_manager.settings
+
+
+@router.post(
+    "/teleop/settings",
+    response_model=StatusResponse,
+    summary="Update Teleop Settings",
+    description="Update teleoperation settings such as VR scaling factor.",
+)
+async def update_teleop_settings(
+    settings: TeleopSettingsRequest,
+    teleop_manager: TeleopManager = Depends(get_teleop_manager),
+) -> StatusResponse:
+    """
+    Update teleoperation settings.
+    """
+    settings_dict = settings.model_dump()
+
+    for attr_name, value in settings_dict.items():
+        if hasattr(teleop_manager, attr_name):
+            setattr(teleop_manager, attr_name, value)
+        else:
+            logger.warning(
+                f"Attempted to set non-existent attribute '{attr_name}' on TeleopManager"
+            )
+
+    return StatusResponse()
