@@ -1,10 +1,9 @@
-import os
-import cv2
-import random
 import base64
+import os
+import random
 import traceback
 from pathlib import Path, PurePath
-from typing import Literal, cast
+from typing import Literal, Union, cast
 
 from fastapi import APIRouter, HTTPException, Request
 from fastapi.responses import FileResponse, HTMLResponse
@@ -17,19 +16,23 @@ from phosphobot.models import (
     AdminSettingsRequest,
     AdminSettingsResponse,
     AdminSettingsTokenResponse,
+    BaseDataset,
     BrowseFilesResponse,
     BrowserFilesRequest,
-    BaseDataset,
     DatasetListResponse,
     DatasetRepairRequest,
     DatasetShuffleRequest,
     DatasetSplitRequest,
     DeleteEpisodeRequest,
+    EpisodesModel,
     HFDownloadDatasetRequest,
+    HFWhoamIResponse,
     HuggingFaceTokenRequest,
+    InfoModel,
     InfoResponse,
-    MergeDatasetsRequest,
     ItemInfo,
+    LeRobotDataset,
+    MergeDatasetsRequest,
     ModelConfigurationRequest,
     ModelConfigurationResponse,
     StatusResponse,
@@ -37,13 +40,11 @@ from phosphobot.models import (
     TrainingInfoResponse,
     VizSettingsResponse,
     WandBTokenRequest,
-    InfoModel,
 )
-from phosphobot.models import EpisodesModel, LeRobotDataset
 from phosphobot.utils import (
     get_hf_token,
-    get_resources_path,
     get_home_app_path,
+    get_resources_path,
     is_running_on_pi,
     login_to_hf,
     parse_hf_username_or_orgid,
@@ -75,7 +76,7 @@ INDEX_PATH = get_resources_path() / "dist" / "index.html"
 @router.get("/train", response_class=HTMLResponse)
 @router.get("/inference", response_class=HTMLResponse)
 @router.get("/", response_class=HTMLResponse)
-async def serve_dashboard(request: Request):
+async def serve_dashboard(request: Request) -> HTMLResponse:
     with open(INDEX_PATH.resolve(), "r") as f:
         content = f.read()
     return HTMLResponse(
@@ -84,7 +85,7 @@ async def serve_dashboard(request: Request):
 
 
 @router.get("/admin/settings", response_model=AdminSettingsResponse)
-async def get_admin_settings():
+async def get_admin_settings() -> AdminSettingsResponse:
     return AdminSettingsResponse(
         dataset_name=config.DEFAULT_DATASET_NAME,
         freq=config.DEFAULT_FREQ,
@@ -98,7 +99,7 @@ async def get_admin_settings():
 
 
 @router.post("/admin/settings/tokens", response_model=AdminSettingsTokenResponse)
-async def get_admin_settings_token():
+async def get_admin_settings_token() -> AdminSettingsTokenResponse:
     # Return bool if the token is set and valid
     return AdminSettingsTokenResponse(
         huggingface=login_to_hf(revalidate=False),
@@ -107,7 +108,7 @@ async def get_admin_settings_token():
 
 
 @router.get("/viz/settings", response_model=VizSettingsResponse)
-def get_viz_settings(request: Request):
+def get_viz_settings(request: Request) -> VizSettingsResponse:
     """
     Page with an overview of the connected cameras. Open this page in the chrome browser.
     """
@@ -186,8 +187,8 @@ def list_directory_items(path: str, root_dir: str = "") -> list[ItemInfo]:
     return items_info
 
 
-@router.post("/files", response_model=BrowseFilesResponse)
-async def files(query: BrowserFilesRequest):
+@router.post("/files", response_model=None)
+async def files(query: BrowserFilesRequest) -> Union[BrowseFilesResponse, FileResponse]:
     # Record page view
     safe_path = sanitize_path(query.path)
     root = Path(ROOT_DIR)
@@ -238,14 +239,12 @@ async def files(query: BrowserFilesRequest):
     return response_data
 
 
-@router.post("/admin/huggingface/whoami")
-async def whoami():
+@router.post("/admin/huggingface/whoami", response_model=HFWhoamIResponse)
+async def whoami() -> HFWhoamIResponse:
     # Fetch the Hugging Face token and try to run whoami to return the person id
     token_path = str(get_home_app_path()) + "/huggingface.token"
     if not os.path.exists(token_path):
-        return {
-            "status": "error",
-        }
+        return HFWhoamIResponse(status="error")
     try:
         with open(token_path, "r") as token_file:
             token = token_file.read().strip()
@@ -255,36 +254,32 @@ async def whoami():
         username_or_orgid = parse_hf_username_or_orgid(user_info)
         logger.info(f"Token is valid for {username_or_orgid}")
         if username_or_orgid is None:
-            return {
-                "status": "error",
-            }
-        return {
-            "status": "success",
-            "username": username_or_orgid,
-        }
+            return HFWhoamIResponse(status="error")
+        return HFWhoamIResponse(
+            status="ok",
+            username=username_or_orgid,
+        )
     except Exception:
-        return {
-            "status": "error",
-        }
+        return HFWhoamIResponse(status="error")
 
 
-@router.post("/admin/huggingface")
-async def submit_token(query: HuggingFaceTokenRequest):
+@router.post("/admin/huggingface", response_model=StatusResponse)
+async def submit_token(query: HuggingFaceTokenRequest) -> StatusResponse:
     if query.token == "":
         # Delete the token file if the token is empty
         file_path = str(get_home_app_path()) + "/huggingface.token"
         if os.path.exists(file_path):
             try:
                 os.remove(file_path)
-                return {
-                    "status": "success",
-                    "message": "The file huggingface.token has been removed.",
-                }
+                return StatusResponse(
+                    status="ok",
+                    message="The file huggingface.token has been removed.",
+                )
             except Exception as e:
-                return {
-                    "status": "error",
-                    "message": f"Error removing Hugging Face token: {str(e)}",
-                }
+                return StatusResponse(
+                    status="error",
+                    message=f"Error removing Hugging Face token: {str(e)}",
+                )
 
     # Try first to connect to HuggingFace with the token
     try:
@@ -293,15 +288,15 @@ async def submit_token(query: HuggingFaceTokenRequest):
         # Get the username or org that has write access
         username_or_orgid = parse_hf_username_or_orgid(user_info)
         if username_or_orgid is None:
-            return {
-                "status": "error",
-                "message": "The token does not have write access to any repository. Please add 'Write access to content/settings' in the token scope.",
-            }
+            return StatusResponse(
+                status="error",
+                message="The token does not have write access to any repository. Please add 'Write access to content/settings' in the token scope.",
+            )
     except Exception as e:
-        return {
-            "status": "error",
-            "message": f"Error connecting to Hugging Face: {str(e)}",
-        }
+        return StatusResponse(
+            status="error",
+            message=f"Error connecting to Hugging Face: {str(e)}",
+        )
 
     # Define the file path where the token will be saved
     file_path = str(get_home_app_path()) + "/huggingface.token"
@@ -317,15 +312,20 @@ async def submit_token(query: HuggingFaceTokenRequest):
         # Auth again to HuggingFace
         login_to_hf()
 
-        return {"status": "success", "message": "Token saved successfully!"}
+        return StatusResponse(
+            status="ok",
+            message="Token saved successfully!",
+        )
 
     except Exception as e:
-        logger.error(f"Error saving token: {str(e)}")
-        return {"status": "error", "message": f"Error saving token: {str(e)}"}
+        return StatusResponse(
+            status="error",
+            message=f"Error saving token: {str(e)}",
+        )
 
 
-@router.post("/admin/wandb")
-async def submit_wandb_token(query: WandBTokenRequest):
+@router.post("/admin/wandb", response_model=StatusResponse)
+async def submit_wandb_token(query: WandBTokenRequest) -> StatusResponse:
     # For now, we don't perform any check on the token
     # TODO: make sure an invalid token won't crash the training process
 
@@ -337,21 +337,21 @@ async def submit_wandb_token(query: WandBTokenRequest):
         if os.path.exists(file_path):
             try:
                 os.remove(file_path)
-                return {
-                    "status": "success",
-                    "message": "The file wandb.token has been removed.",
-                }
+                return StatusResponse(
+                    status="ok",
+                    message="The file wandb.token has been removed.",
+                )
             except Exception as e:
-                return {
-                    "status": "error",
-                    "message": f"Error removing WandB token: {str(e)}",
-                }
+                return StatusResponse(
+                    status="error",
+                    message=f"Error removing WandB token: {str(e)}",
+                )
 
     if len(query.token) != 40:
-        return {
-            "status": "error",
-            "message": "Wrong token, make sure to copy/paste the 40 characters long token at https://wandb.ai/authorize",
-        }
+        return StatusResponse(
+            status="error",
+            message="Wrong token, make sure to copy/paste the 40 characters long token at https://wandb.ai/authorize",
+        )
 
     try:
         # Open the file in write mode and save the token
@@ -362,30 +362,31 @@ async def submit_wandb_token(query: WandBTokenRequest):
         # Change config
         logger.info("Token saved successfully!")
 
-        return {"status": "success", "message": "Token saved successfully!"}
+        return StatusResponse(status="ok", message="Token saved successfully!")
 
     except Exception as e:
-        logger.error(f"Error saving token: {str(e)}")
-        return {
-            "status": "error",
-            "message": f"Error saving token: {str(e)}",
-        }
+        raise HTTPException(status_code=400, detail=f"Error saving token: {str(e)}")
 
 
-@router.post("/admin/form/usersettings")
-async def submit_user_settings(user_settings: AdminSettingsRequest):
+@router.post("/admin/form/usersettings", response_model=StatusResponse)
+async def submit_user_settings(user_settings: AdminSettingsRequest) -> StatusResponse:
     try:
         logger.debug(f"User Settings submission: {user_settings.model_dump()}")
         config.save_user_settings(user_settings.model_dump())
 
-        return {"status": "success", "message": "User settings saved successfully!"}
+        return StatusResponse(
+            status="ok",
+            message="User settings saved successfully!",
+        )
     except Exception as e:
-        logger.error(f"User Settings submission error: {e}")
-        return {"status": "error", "message": f"Error saving user settings: {str(e)}"}
+        raise HTTPException(
+            status_code=400,
+            detail=f"Error saving user settings: {str(e)}",
+        )
 
 
-@router.post("/dataset/delete")
-async def delete_dataset(request: Request, path: str):
+@router.post("/dataset/delete", response_model=StatusResponse)
+async def delete_dataset(request: Request, path: str) -> StatusResponse:
     dataset_path = os.path.join(ROOT_DIR, path)
     # Check if the path exists and is a directory
     if not os.path.exists(dataset_path) or not os.path.isdir(dataset_path):
@@ -440,6 +441,8 @@ async def get_dataset_info(path: str) -> InfoResponse:
             status="error",
         )
 
+    import cv2
+
     image_frames = {}
     for key in info.features.observation_images.keys():
         # Extract first frame from first video file
@@ -474,7 +477,7 @@ async def get_dataset_info(path: str) -> InfoResponse:
 
 
 @router.post("/dataset/merge")
-async def merge_datasets(merge_request: MergeDatasetsRequest):
+async def merge_datasets(merge_request: MergeDatasetsRequest) -> StatusResponse:
     """
     Merge two datasets into one.
     """
@@ -588,7 +591,7 @@ async def merge_datasets(merge_request: MergeDatasetsRequest):
 
 
 @router.post("/dataset/list", response_model=DatasetListResponse)
-async def list_datasets():
+async def list_datasets() -> DatasetListResponse:
     """
     List all datasets that are both in Hugging Face and locally.
     """
@@ -636,7 +639,7 @@ async def list_datasets():
 
 
 @router.post("/episode/delete")
-async def delete_episode(query: DeleteEpisodeRequest):
+async def delete_episode(query: DeleteEpisodeRequest) -> StatusResponse:
     """
     Delete an episode from the dataset.
     Parameters:
@@ -667,7 +670,7 @@ async def delete_episode(query: DeleteEpisodeRequest):
 
 
 @router.post("/dataset/sync")
-async def sync_dataset(path: str):
+async def sync_dataset(path: str) -> StatusResponse:
     # Extract dataset name et huggingface repo id from the path
     dataset = BaseDataset(path=os.path.join(ROOT_DIR, path))
 
@@ -678,7 +681,10 @@ async def sync_dataset(path: str):
 
 
 @router.get("/dataset/download")
-async def download_folder(folder_path: str):
+async def download_folder(folder_path: str) -> FileResponse:
+    """
+    Download a folder as a ZIP file.
+    """
     # Construct the full path
     full_path = os.path.join(ROOT_DIR, folder_path)
 
@@ -735,7 +741,8 @@ async def get_training_info(
     request: TrainingInfoRequest,
 ) -> TrainingInfoResponse:
     """
-    Fetch the info.json from the model repo and return the training info.
+    - Fetch the info.json from the model repo and return the training info.
+    - If the model type is "custom", return a custom command to run the training.
     """
     if request.model_type == "custom":
         return TrainingInfoResponse(
@@ -745,6 +752,15 @@ async def get_training_info(
             },
         )
 
+    if request.model_id is None or request.model_id.strip() == "":
+        raise HTTPException(
+            status_code=400,
+            detail="Model ID is required to fetch training info.",
+        )
+    logger.debug(
+        f"Fetching training info for model {request.model_id} of type {request.model_type}"
+    )
+
     try:
         token_path = str(get_home_app_path()) + "/huggingface.token"
 
@@ -788,10 +804,24 @@ async def get_training_info(
         random_suffix = "".join(
             random.choices("abcdefghijklmnopqrstuvwxyz0123456789", k=5)
         )
+
+        # Determine private mode from config
+        private_mode = config.DEFAULT_HF_PRIVATE_MODE
+
+        # Set model name based on private mode
+        if private_mode:
+            # Use user's namespace for private training
+            model_name = f"{username_or_orgid}/{request.model_type}-{request.model_id.split('/')[1]}-{random_suffix}"
+        else:
+            # Use phospho-app namespace for public training
+            model_name = f"phospho-app/{username_or_orgid}-{request.model_type}-{request.model_id.split('/')[1]}-{random_suffix}"
+
         training_response = TrainingRequest(
             model_type=request.model_type,
             dataset_name=request.model_id,
-            model_name=f"phospho-app/{username_or_orgid}-{request.model_type}-{request.model_id.split('/')[1]}-{random_suffix}",
+            model_name=model_name,
+            private_mode=private_mode,
+            user_hf_token=None,
         )
         # Replace the fields in training_response with the values from training_params dict
         if training_response.training_params is not None:
@@ -799,19 +829,22 @@ async def get_training_info(
                 if hasattr(training_response.training_params, key):
                     setattr(training_response.training_params, key, value)
 
+        # Prepare the training body with masked token
+        training_body = training_response.model_dump(
+            exclude={
+                "wandb_api_key": True,
+                "custom_command": True,
+                "training_params": {
+                    "data_dir": True,
+                    "output_dir": True,
+                    "validation_data_dir": True,
+                },
+            }
+        )
+
         return TrainingInfoResponse(
             status="ok",
-            training_body=training_response.model_dump(
-                exclude={
-                    "wandb_api_key": True,
-                    "custom_command": True,
-                    "training_params": {
-                        "path_to_gr00t_repo": True,
-                        "data_dir": True,
-                        "output_dir": True,
-                    },
-                }
-            ),
+            training_body=training_body,
         )
     except Exception as e:
         logger.warning(f"Error fetching training info: {e}")
@@ -822,9 +855,7 @@ async def get_training_info(
 
 
 @router.post("/dataset/hf_download")
-async def hf_download_dataset(
-    query: HFDownloadDatasetRequest,
-) -> StatusResponse:
+async def hf_download_dataset(query: HFDownloadDatasetRequest) -> StatusResponse:
     if os.path.exists(os.path.join(ROOT_DIR, "lerobot_v2.1", query.dataset_name)):
         return StatusResponse(
             status="error",
@@ -894,7 +925,7 @@ async def hf_download_dataset(
 
 
 @router.post("/dataset/repair", response_model=StatusResponse)
-async def repair_dataset(query: DatasetRepairRequest):
+async def repair_dataset(query: DatasetRepairRequest) -> StatusResponse:
     """
     Repair a dataset by removing any corrupted files.
     For now, this only works for parquets files.
@@ -922,7 +953,7 @@ async def repair_dataset(query: DatasetRepairRequest):
 
 
 @router.post("/dataset/split", response_model=StatusResponse)
-async def split_dataset(query: DatasetSplitRequest):
+async def split_dataset(query: DatasetSplitRequest) -> StatusResponse:
     """
     Split a dataset into two datasets.
     Used for creating training and validation datasets.
@@ -960,7 +991,7 @@ async def split_dataset(query: DatasetSplitRequest):
 
 
 @router.post("/dataset/shuffle", response_model=StatusResponse)
-async def shuffle_dataset(query: DatasetShuffleRequest):
+async def shuffle_dataset(query: DatasetShuffleRequest) -> StatusResponse:
     """
     Shuffle a dataset in place.
     """
@@ -978,29 +1009,11 @@ async def shuffle_dataset(query: DatasetShuffleRequest):
             message="You can only shuffle datasets of type v2.1",
         )
 
-    # Shuffle the dataset
+    # Create a dataset object
     dataset = LeRobotDataset(path=dataset_path, enforce_path=True)
 
-    # Name of the new dataset after shuffling
-    new_dataset_name = f"{query.dataset_path}_shuffled"
-    while os.path.exists(os.path.join(ROOT_DIR, new_dataset_name)):
-        # Find if a int suffix is already present
-        if new_dataset_name.endswith("_shuffled"):
-            # If it ends with _shuffled, we can add a number
-            new_dataset_name += "_1"
-        else:
-            # Otherwise, parse the int and increment it
-            try:
-                suffix = int(new_dataset_name.split("_")[-1])
-                new_dataset_name = (
-                    "_".join(new_dataset_name.split("_")[:-1]) + f"_{suffix + 1}"
-                )
-            except ValueError:
-                # If the suffix is not an int, we just add _1
-                new_dataset_name += "_1"
-
     try:
-        dataset.shuffle_dataset(new_dataset_name=new_dataset_name)
+        dataset.shuffle_dataset()
     except Exception as e:
         logger.warning(f"Error shuffling dataset: {e}")
         return StatusResponse(
