@@ -1,18 +1,19 @@
+import asyncio
 import atexit
 import json
 import os
-import asyncio
 from abc import abstractmethod
-from typing import List, Literal, Optional, Union
+from typing import Any, Dict, List, Literal, Optional, Tuple, Union
 
-from fastapi import HTTPException
 import numpy as np
+from fastapi import HTTPException
 from loguru import logger
+from scipy.spatial.transform import Rotation as R  # type: ignore
+
 from phosphobot.configs import config as cfg
+from phosphobot.hardware import get_sim
 from phosphobot.models import BaseRobot, BaseRobotConfig, BaseRobotInfo, Temperature
 from phosphobot.models.lerobot_dataset import FeatureDetails
-from phosphobot.hardware import get_sim
-from scipy.spatial.transform import Rotation as R  # type: ignore
 from phosphobot.utils import (
     euler_from_quaternion,
     get_resources_path,
@@ -24,15 +25,15 @@ class AxisRobot:
     Used to place the robots in a grid.
     """
 
-    def __init__(self):
+    def __init__(self) -> None:
         # Create a grid of (x, y, 0) positions with a step of 0.4
-        self.grid = []
+        self.grid: List[List[float]] = []
         for x in np.arange(0, 10, 1):
             for y in np.arange(0, 10, 1):
-                self.grid.append([x, y, 0])
+                self.grid.append([float(x), float(y), 0])
         self.grid_index = 0
 
-    def new_position(self):
+    def new_position(self) -> List[float]:
         if self.grid_index >= len(self.grid):
             self.grid_index = 0
         axis = self.grid[self.grid_index]
@@ -63,14 +64,15 @@ class BaseManipulator(BaseRobot):
     # They are in the same order as the joint links in the URDF file
     SERVO_IDS: List[int]
 
-    CALIBRATION_POSITION: list[float]  # same size as SERVO_IDS
-    SLEEP_POSITION: list[float] | None = None
+    CALIBRATION_POSITION: List[float]  # same size as SERVO_IDS
+    SLEEP_POSITION: Optional[List[float]] = None
+    time_to_sleep: float = 0.7 # seconds to wait after moving to sleep position
     RESOLUTION: int
     # The effector is the gripper
     END_EFFECTOR_LINK_INDEX: int
 
     # calibration config: offsets, signs, pid values
-    config: BaseRobotConfig | None = None
+    config: Optional[BaseRobotConfig] = None
 
     # status variables
     is_connected: bool = False
@@ -87,9 +89,9 @@ class BaseManipulator(BaseRobot):
     calibration_max_steps: int = 3
 
     # (x, y, z) position of the robot in the simulation in meters
-    initial_orientation_rad: np.ndarray | None = None
+    initial_orientation_rad: Optional[np.ndarray] = None
     # (rx, ry, rz) orientation of the robot in the simulation
-    initial_position: np.ndarray | None = None
+    initial_position: Optional[np.ndarray] = None
 
     @abstractmethod
     def enable_torque(self) -> None:
@@ -110,7 +112,7 @@ class BaseManipulator(BaseRobot):
         raise NotImplementedError("The robot enable torque must be implemented.")
 
     @abstractmethod
-    def read_motor_torque(self, servo_id: int) -> float | None:
+    def read_motor_torque(self, servo_id: int) -> Optional[float]:
         """
         Read the torque of a motor
 
@@ -119,7 +121,7 @@ class BaseManipulator(BaseRobot):
         raise NotImplementedError("The robot enable torque must be implemented.")
 
     @abstractmethod
-    def read_motor_voltage(self, servo_id: int) -> float | None:
+    def read_motor_voltage(self, servo_id: int) -> Optional[float]:
         """
         Read the voltage of a motor
 
@@ -127,7 +129,7 @@ class BaseManipulator(BaseRobot):
         """
         raise NotImplementedError("The robot enable torque must be implemented.")
 
-    def read_motor_temperature(self, servo_id: int) -> tuple[float, float] | None:
+    def read_motor_temperature(self, servo_id: int) -> Optional[Tuple[float, float]]:
         """
         Read the temperature of a motor
         raise: Exception if the routine has not been implemented
@@ -148,7 +150,7 @@ class BaseManipulator(BaseRobot):
         )
 
     @abstractmethod
-    def write_motor_position(self, servo_id: int, units: int, **kwargs) -> None:
+    def write_motor_position(self, servo_id: int, units: int, **kwargs: Any) -> None:
         """
         Move the motor to the specified position.
 
@@ -160,14 +162,14 @@ class BaseManipulator(BaseRobot):
         raise NotImplementedError("The robot write motor position must be implemented.")
 
     @abstractmethod
-    def read_motor_position(self, servo_id: int, **kwargs) -> int | None:
+    def read_motor_position(self, servo_id: int, **kwargs: Any) -> Optional[int]:
         """
         Read the position of the motor. This should return the position in motor units.
         """
         raise NotImplementedError("The robot read motor position must be implemented.")
 
     @abstractmethod
-    def calibrate_motors(self, **kwargs) -> None:
+    def calibrate_motors(self, **kwargs: Any) -> None:
         """
         This is called during the calibration phase of the robot.
         It sets the offset of all motors to self.RESOLUTION/2.
@@ -190,14 +192,14 @@ class BaseManipulator(BaseRobot):
 
     def __init__(
         self,
-        device_name: str | None = None,
-        serial_id: str | None = None,
+        device_name: Optional[str] = None,
+        serial_id: Optional[str] = None,
         only_simulation: bool = False,
         reset_simulation_bool: bool = False,
-        axis: List[float] | None = None,
+        axis: Optional[List[float]] = None,
         add_debug_lines: bool = False,
         show_debug_link_indices: bool = False,
-        **kwargs: Optional[dict[str, str]],
+        **kwargs: Optional[Dict[str, str]],
     ):
         """
         Args:
@@ -228,17 +230,13 @@ class BaseManipulator(BaseRobot):
             self.sim.reset()
 
         logger.info(f"Loading URDF file: {self.URDF_FILE_PATH}")
-        if not os.path.exists(self.URDF_FILE_PATH):
-            raise FileNotFoundError(
-                f"URDF file not found: {self.URDF_FILE_PATH}\nCurrent path: {os.getcwd()}"
-            )
         self.p_robot_id, num_joints, actuated_joints = self.sim.load_urdf(
             urdf_path=self.URDF_FILE_PATH,
             axis=axis,
             axis_orientation=self.AXIS_ORIENTATION,
             use_fixed_base=True,
         )
-
+        self.num_joints = num_joints
         self.actuated_joints = actuated_joints
 
         # Infer SERVO_IDS and CALIBRATION_POSITION from the actuated joints if not set
@@ -248,12 +246,12 @@ class BaseManipulator(BaseRobot):
                 f"{self.__class__.__name__}.SERVO_IDS not set, using default: {self.SERVO_IDS}"
             )
         if not hasattr(self, "CALIBRATION_POSITION"):
-            self.CALIBRATION_POSITION = [0] * len(self.SERVO_IDS)
+            self.CALIBRATION_POSITION = [0.0] * len(self.SERVO_IDS)
             logger.warning(
                 f"{self.__class__.__name__}.CALIBRATION_POSITION not set, using default: {self.CALIBRATION_POSITION}"
             )
         if not hasattr(self, "SLEEP_POSITION"):
-            self.SLEEP_POSITION = [0] * len(self.SERVO_IDS)
+            self.SLEEP_POSITION = [0.0] * len(self.SERVO_IDS)
             logger.warning(
                 f"{self.__class__.__name__}.SLEEP_POSITION not set, using default: {self.SLEEP_POSITION}"
             )
@@ -318,7 +316,7 @@ class BaseManipulator(BaseRobot):
         else:
             self.config = None
 
-    def __del__(self):
+    def __del__(self) -> None:
         try:
             loop = asyncio.get_event_loop()
             if loop.is_running():
@@ -389,9 +387,9 @@ class BaseManipulator(BaseRobot):
             logger.warning("None torque value for gripper motor ")
             current_gripper_torque = np.int32(0)
 
-        return current_gripper_torque
+        return np.int32(current_gripper_torque)
 
-    async def move_to_initial_position(self, open_gripper: bool = False):
+    async def move_to_initial_position(self, open_gripper: bool = False) -> None:
         """
         Move the robot to its initial position.
         """
@@ -408,7 +406,7 @@ class BaseManipulator(BaseRobot):
             self.initial_orientation_rad,
         ) = self.forward_kinematics()
 
-    async def move_to_sleep(self):
+    async def move_to_sleep(self) -> None:
         """
         Move the robot to its sleep position and disable torque.
         """
@@ -420,7 +418,7 @@ class BaseManipulator(BaseRobot):
                     )
                 except Exception:
                     pass
-            await asyncio.sleep(0.7)
+            await asyncio.sleep(self.time_to_sleep)
             self.disable_torque()
             await asyncio.sleep(0.1)
 
@@ -588,7 +586,7 @@ class BaseManipulator(BaseRobot):
 
     def forward_kinematics(
         self, sync_robot_pos: bool = False
-    ) -> tuple[np.ndarray, np.ndarray]:
+    ) -> Tuple[np.ndarray, np.ndarray]:
         """
         Compute the forward kinematics of the robot
         Returns cartesian position and orientation in radians
@@ -634,7 +632,7 @@ class BaseManipulator(BaseRobot):
 
     def get_end_effector_state(
         self, sync: bool = False
-    ) -> tuple[np.ndarray, np.ndarray, float]:
+    ) -> Tuple[np.ndarray, np.ndarray, float]:
         """
         Return the position and orientation in radians of the end effector and the gripper opening value.
         The gripper opening value between 0 and 1.
@@ -749,12 +747,12 @@ class BaseManipulator(BaseRobot):
                 # Normalize the angles to [min_value, max_value]
                 output_position = min_value + (max_value - min_value) * (
                     output_position_rad + np.pi
-                ) / (2 * np.pi)
+                ) / (2 * np.pi)  # type: ignore
 
             elif source_unit == "rad":
                 output_position = min_value + (max_value - min_value) * (
                     output_position + np.pi
-                ) / (2 * np.pi)
+                ) / (2 * np.pi)  # type: ignore
         else:
             raise ValueError(
                 f"Invalid unit: {unit}. Must be one of ['rad', 'motor_units', 'degrees']"
@@ -844,8 +842,8 @@ class BaseManipulator(BaseRobot):
         angles: List[float],
         unit: Literal["rad", "motor_units", "degrees", "other"],
         joints_ids: Optional[List[int]] = None,
-        min_value: float | None = None,
-        max_value: float | None = None,
+        min_value: Optional[float] = None,
+        max_value: Optional[float] = None,
     ) -> None:
         """
         Move the robot's joints to the specified angles.
@@ -933,10 +931,8 @@ class BaseManipulator(BaseRobot):
     async def move_robot_absolute(
         self,
         target_position: np.ndarray,  # cartesian np.array
-        target_orientation_rad: np.ndarray | None,  # rad np.array
-        interpolate_trajectory: bool = False,
-        steps: int = 10,
-        **kwargs,
+        target_orientation_rad: Optional[np.ndarray],  # rad np.array
+        **kwargs: Dict[str, Any],
     ) -> None:
         """
         Move the robot to the absolute position and orientation.
@@ -991,10 +987,7 @@ class BaseManipulator(BaseRobot):
         )
 
         self.is_moving = True
-        if not interpolate_trajectory:
-            self.set_motors_positions(goal_q_robot_rad)
-        else:
-            raise NotImplementedError("Interpolation not implemented yet")
+        self.set_motors_positions(goal_q_robot_rad)
         self.is_moving = False
 
         # reset gripping status when going to init position
@@ -1007,7 +1000,7 @@ class BaseManipulator(BaseRobot):
         self.sim.set_joints_states(
             robot_id=self.p_robot_id,
             joint_indices=self.actuated_joints,
-            target_positions=joints,
+            target_positions=list(joints),
         )
         # Update the simulation
         self.sim.step()
@@ -1068,11 +1061,7 @@ class BaseManipulator(BaseRobot):
         self.SLEEP_POSITION = None
         self.config = None
 
-    def control_gripper(
-        self,
-        open_command: float,  # Should be between 0 and 1
-        **kwargs,
-    ) -> None:
+    def control_gripper(self, open_command: float, **kwargs: Any) -> None:
         """
         Open or close the gripper until object is gripped.
         open_command: 0 to close, 1 to open
@@ -1095,6 +1084,14 @@ class BaseManipulator(BaseRobot):
             self.write_gripper_command(open_command_clipped)
         self.closing_gripper_value = open_command_clipped
 
+        # Update simulation only if the object has not been gripped:
+        self.move_gripper_in_sim(open=open_command_clipped)
+    
+
+    def move_gripper_in_sim(self, open: float) -> None:
+        """
+        Move the gripper in the simulation.
+        """
         ## Simulation side
         # Since last motor ID might not be equal to the number of motors ( due to some shadowed motors)
         # We extract last motor calibration data for the gripper:
@@ -1107,20 +1104,21 @@ class BaseManipulator(BaseRobot):
         else:
             open_position = self.upper_joint_limits[-1]
 
-        # Update simulation only if the object has not been gripped:
         if not self.is_object_gripped:
             self.sim.set_joints_states(
-                robot_id=self.p_robot_id,
-                joint_indices=[self.GRIPPER_JOINT_INDEX],
-                target_positions=[
-                    close_position
-                    + (open_position - close_position) * open_command_clipped
-                ],
-            )
+                    robot_id=self.p_robot_id,
+                    joint_indices=[self.GRIPPER_JOINT_INDEX],
+                    target_positions=[
+                        close_position
+                        + (open_position - close_position) * open
+                    ],
+                )
 
     def get_observation(
-        self, do_forward: bool = False
-    ) -> tuple[np.ndarray, np.ndarray]:
+        self,
+        source: Literal["sim", "robot"],
+        do_forward: bool = False,
+    ) -> Tuple[np.ndarray, np.ndarray]:
         """
         Get the observation of the robot.
 
@@ -1132,51 +1130,26 @@ class BaseManipulator(BaseRobot):
             - joints_position: np.array joints position of the robot
         """
 
-        from phosphobot.endpoints.control import (
-            signal_ai_control,
-            signal_leader_follower,
-            signal_vr_control,
-        )
-
-        if (
-            signal_ai_control.is_in_loop()
-            or signal_leader_follower.is_in_loop()
-            or signal_vr_control.is_in_loop()
-        ):
-            joints_position = self.read_joints_position(unit="rad", source="sim")
-        else:
-            joints_position = self.read_joints_position(unit="rad", source="robot")
+        joints_position = self.read_joints_position(unit="rad", source=source)
 
         if do_forward:
             effector_position, effector_orientation_euler_rad = (
                 self.forward_kinematics()
             )
-            state = np.concatenate((effector_position, effector_orientation_euler_rad))
+            state = np.concatenate(
+                (
+                    effector_position,
+                    effector_orientation_euler_rad,
+                    [joints_position[-1]],
+                )
+            )
         else:
             # Skip forward kinematics and return nan values
-            state = np.full(6, np.nan)
+            # This is size 7 for [x, y, z, rx, ry, rz, gripper]
+            state = np.full(7, np.nan)
 
         return state, joints_position
 
-    def mimick_simu_to_robot(self):
-        """
-        Update simulation base on leader robot reading of joint position.
-        """
-        joints_position = self.read_joints_position(unit="rad")
-        gripper_command = self.read_gripper_command()
-
-        # Update simulation
-        # this take into account the leader that has less joints
-        logger.debug(f"Moving to position: {joints_position}")
-        self.sim.set_joints_states(
-            robot_id=self.p_robot_id,
-            joint_indices=self.actuated_joints,
-            target_positions=joints_position,
-        )
-        # Update the simulation
-        self.sim.step()
-        self.control_gripper(gripper_command)
-
     def get_info_for_dataset(self) -> BaseRobotInfo:
         """
         Get information about a robot.
@@ -1201,7 +1174,7 @@ class BaseManipulator(BaseRobot):
             ),
         )
 
-    def current_voltage(self) -> np.ndarray | None:
+    def current_voltage(self) -> Optional[np.ndarray]:
         """
         Read the current voltage u of the joints of the robot.
 
@@ -1221,12 +1194,11 @@ class BaseManipulator(BaseRobot):
         # If the robot is not connected, error raised
         return None
 
-    def current_temperature(self) -> List[Temperature] | None:
+    def current_temperature(self) -> Optional[List[Temperature]]:
         """
         Read the current and maximum temperature of the joints of the robot.
         Returns:
-            List[Temperature] | None: A list of Temperature objects, one for each joint,
-                                    or None if the robot is not connected.
+            A list of Temperature objects, one for each joint, or None if the robot is not connected.
         """
         if self.is_connected:
             temperatures = []
@@ -1301,7 +1273,7 @@ class BaseManipulator(BaseRobot):
 
         return current_torque
 
-    def update_object_gripping_status(self):
+    def update_object_gripping_status(self) -> None:
         """
         Based on the torque value, update the object gripping status.
 
@@ -1348,7 +1320,7 @@ class BaseMobileRobot(BaseRobot):
     def __init__(
         self,
         only_simulation: bool = False,
-    ):
+    ) -> None:
         if not only_simulation:
             # Register the disconnect method to be called on exit
             atexit.register(self.move_to_sleep_sync)
