@@ -1,6 +1,6 @@
 import asyncio
 import time
-from typing import Literal, Optional, List
+from typing import Any, Dict, List, Literal, Optional, Tuple, cast
 
 import numpy as np
 from loguru import logger
@@ -10,8 +10,8 @@ from phosphobot.configs import SimulationMode, config
 from phosphobot.control_signal import ControlSignal
 from phosphobot.hardware.base import BaseManipulator
 from phosphobot.hardware.motors.feetech import FeetechMotorsBus  # type: ignore
-from phosphobot.utils import get_resources_path
 from phosphobot.models import RobotConfigStatus
+from phosphobot.utils import get_resources_path
 
 
 class SO100Hardware(BaseManipulator):
@@ -23,20 +23,11 @@ class SO100Hardware(BaseManipulator):
 
     AXIS_ORIENTATION = [0, 0, 1, 1]
 
-    # Control commands (refer to the Feetech SCServo manual)
-    TORQUE_ENABLE = 0x01
-    TORQUE_DISABLE = 0
-
-    TORQUE_ADDRESS = 0x40
-
-    COMMAND_WRITE = 0x03
-    COMMAND_READ = 0x02
-
     END_EFFECTOR_LINK_INDEX = 4
     GRIPPER_JOINT_INDEX = 5
 
-    # Dynamixel settings
-    motors = {
+    # Feetech settings
+    motors: Dict[str, List[object]] = {
         # name: (index, model)
         "shoulder_pan": [1, "sts3215"],
         "shoulder_lift": [2, "sts3215"],
@@ -74,11 +65,12 @@ class SO100Hardware(BaseManipulator):
     _gravity_task: Optional[asyncio.Task] = None
 
     @property
-    def servo_id_to_motor_name(self):
-        return {v[0]: k for k, v in self.motors.items()}
+    def servo_id_to_motor_name(self) -> Dict[int, str]:
+        output: Dict[int, str] = {cast(int, v[0]): k for k, v in self.motors.items()}
+        return output
 
     @classmethod
-    def from_port(cls, port: ListPortInfo, **kwargs) -> Optional["SO100Hardware"]:
+    def from_port(cls, port: ListPortInfo, **kwargs: Any) -> Optional["SO100Hardware"]:
         """
         Detect if the device is a SO-100 robot.π
         """
@@ -103,9 +95,10 @@ class SO100Hardware(BaseManipulator):
         # Create serial connection
         self.motors_bus = FeetechMotorsBus(port=self.device_name, motors=self.motors)
         self.motors_bus.connect()
-        self.is_connected = True
         self.init_config()
         self._max_temperature_cache: dict = {}
+        self.is_connected = True
+
 
     def disconnect(self) -> None:
         """
@@ -192,7 +185,30 @@ class SO100Hardware(BaseManipulator):
                 "Motors torque is disabled. Motors must have torque enabled to change PID coefficients. Enable torque first."
             )
 
-    def update_motor_errors(self):
+    def _get_pid_gains_motor(self, servo_id: int) -> Optional[Tuple[int, int, int]]:
+        """
+        Get the PID gains for the Feetech servo.
+
+        :servo_id: Joint ID (0-6)
+        :return: Tuple of (p_gain, i_gain, d_gain)
+        """
+        try:
+            p_gain = self.motors_bus.read(
+                "P_Coefficient", motor_names=self.servo_id_to_motor_name[servo_id]
+            )
+            i_gain = self.motors_bus.read(
+                "I_Coefficient", motor_names=self.servo_id_to_motor_name[servo_id]
+            )
+            d_gain = self.motors_bus.read(
+                "D_Coefficient", motor_names=self.servo_id_to_motor_name[servo_id]
+            )
+            return int(p_gain), int(i_gain), int(d_gain)
+        except Exception as e:
+            logger.warning(f"Error reading PID gains: {e}")
+            self.update_motor_errors()
+            return None
+
+    def update_motor_errors(self) -> None:
         """
         Every time a motor communication error is detected, increment the error counter.
         If the counter reaches a certain threshold, disconnect the robot.
@@ -205,7 +221,7 @@ class SO100Hardware(BaseManipulator):
             logger.error("Too many communication errors. Disconnecting robot.")
             self.disconnect()
 
-    def read_motor_position(self, servo_id: int, **kwargs) -> int | None:
+    def read_motor_position(self, servo_id: int, **kwargs: Any) -> Optional[int]:
         """
         Read the position of a Feetech servo.
         """
@@ -223,7 +239,7 @@ class SO100Hardware(BaseManipulator):
             self.update_motor_errors()
             return None
 
-    def write_motor_position(self, servo_id: int, units: int, **kwargs) -> None:
+    def write_motor_position(self, servo_id: int, units: int, **kwargs: Any) -> None:
         """
         Write a position to a Feetech servo.
         """
@@ -289,7 +305,7 @@ class SO100Hardware(BaseManipulator):
             return np.ones(6) * np.nan
         return motor_positions
 
-    def read_motor_torque(self, servo_id: int, **kwargs) -> float | None:
+    def read_motor_torque(self, servo_id: int, **kwargs: Any) -> Optional[float]:
         """
         Read the torque of a Feetech servo.
         """
@@ -307,7 +323,7 @@ class SO100Hardware(BaseManipulator):
             self.update_motor_errors()
             return None
 
-    def read_motor_voltage(self, servo_id: int, **kwargs) -> float | None:
+    def read_motor_voltage(self, servo_id: int, **kwargs: Any) -> Optional[float]:
         """
         Read the voltage of a Feetech servo.
         """
@@ -334,8 +350,8 @@ class SO100Hardware(BaseManipulator):
         )
 
     def read_motor_temperature(
-        self, servo_id: int, **kwargs
-    ) -> tuple[float, float] | None:
+        self, servo_id: int, **kwargs: Any
+    ) -> Optional[Tuple[float, float]]:
         """
         Read the temperature of a Feetech servo.
         """
@@ -364,7 +380,7 @@ class SO100Hardware(BaseManipulator):
             return None
 
     def write_group_motor_maximum_temperature(
-        self, maximum_temperature_target: List[int], **kwargs
+        self, maximum_temperature_target: List[int], **kwargs: Any
     ) -> None:
         """
         Write the maximum temperature of all motors of a robot.
@@ -537,7 +553,7 @@ class SO100Hardware(BaseManipulator):
             f"Invalid calibration step: {self.calibration_current_step}, must be between 0 and {self.calibration_max_steps - 1}"
         )
 
-    def calibrate_motors(self, **kwargs) -> None:
+    def calibrate_motors(self, **kwargs: Any) -> None:
         """
         This is called during the calibration phase of the robot.
         It sets the offset of all motors to self.RESOLUTION/2.
@@ -552,7 +568,7 @@ class SO100Hardware(BaseManipulator):
     async def gravity_compensation_loop(
         self,
         control_signal: ControlSignal,
-    ):
+    ) -> None:
         """
         Background task that implements gravity compensation control:
         - Applies gravity compensation to the robot
@@ -600,7 +616,7 @@ class SO100Hardware(BaseManipulator):
 
         # Main control loop
         while control_signal.is_in_loop():
-            start_time = time.time()
+            start_time = time.perf_counter()
 
             # Get leader's current joint positions
             pos_rad = self.read_joints_position(unit="rad")
@@ -623,10 +639,10 @@ class SO100Hardware(BaseManipulator):
 
             # Apply gravity compensation to leader
             theta_des_rad = pos_rad + alpha[:num_joints] * np.array(tau_g)
-            self.write_joint_positions(theta_des_rad, unit="rad")
+            self.write_joint_positions(theta_des_rad.tolist(), unit="rad")
 
             # Maintain loop frequency
-            elapsed = time.time() - start_time
+            elapsed = time.perf_counter() - start_time
             sleep_time = max(0, loop_period - elapsed)
             await asyncio.sleep(sleep_time)
 
