@@ -2,28 +2,30 @@ import asyncio
 import os
 import time
 from concurrent.futures import ThreadPoolExecutor
-from typing import Literal, Optional, List
+from typing import List, Literal, Optional, Tuple, Dict
 
 import numpy as np
 from fastapi import BackgroundTasks, Depends
 from loguru import logger
 
-from phosphobot.camera import AllCameras, get_all_cameras
+from phosphobot.camera import AllCameras, BaseCamera, get_all_cameras
 from phosphobot.configs import config
 from phosphobot.hardware import BaseRobot
-from phosphobot.models import BaseDataset, Observation, Step
 
 # New imports for refactored Episode structure
 from phosphobot.models import (
+    BaseDataset,
     BaseEpisode,
     JsonEpisode,
-    LeRobotEpisode,
     LeRobotDataset,
+    LeRobotEpisode,
+    Observation,
+    Step,
 )
+from phosphobot.rerun_visualizer import RerunVisualizer
 from phosphobot.robot import RobotConnectionManager, get_rcm
 from phosphobot.types import VideoCodecs
 from phosphobot.utils import background_task_log_exceptions, get_home_app_path
-from phosphobot.rerun_visualizer import RerunVisualizer
 
 recorder = None  # Global variable to store the recorder instance
 
@@ -34,7 +36,7 @@ class Recorder:
     is_saving: bool = False
     is_recording: bool = False
     episode: Optional[BaseEpisode] = None  # Link to an Episode instance
-    start_ts: float | None
+    start_ts: Optional[float]
     freq: int  # Stored from start() for use in record_loop
 
     cameras: AllCameras
@@ -84,17 +86,24 @@ class Recorder:
     async def start(
         self,
         background_tasks: BackgroundTasks,
-        robots: list[BaseRobot],  # Can use self.robots or update if new list passed
+        robots: List[BaseRobot],
+        actions_robots_mapping: Dict[int, Literal["sim", "robot"]],
+        observations_robots_mapping: Dict[int, Literal["sim", "robot"]],
         codec: VideoCodecs,
         freq: int,
-        target_size: tuple[int, int] | None,
+        target_size: Optional[Tuple[int, int]],
         dataset_name: str,  # e.g., "my_robot_data"
-        instruction: str | None,
+        instruction: Optional[str],
         episode_format: Literal["json", "lerobot_v2", "lerobot_v2.1"],
-        cameras_ids_to_record: list[int] | None,
+        cameras_ids_to_record: Optional[List[int]],
         use_push_to_hf: bool = True,  # Stored for save_episode to decide
-        branch_path: str | None = None,  # Stored for push_to_hub if initiated from here
+        # Stored for push_to_hub if initiated from here
+        branch_path: Optional[str] = None,
         enable_rerun: bool = False,  # Enable real-time Rerun visualization
+        save_cartesian: bool = False,  # Saves cartesian positions if True (only for robots with simulators)
+        add_metadata: Optional[
+            Dict[str, list]
+        ] = None,  # Additional metadata to save with each step
     ) -> None:
         if target_size is None:
             target_size = (config.DEFAULT_VIDEO_SIZE[0], config.DEFAULT_VIDEO_SIZE[1])
@@ -105,7 +114,9 @@ class Recorder:
             )
             await self.stop()  # Stop does not save, just halts the loop
 
-        self.robots = robots  # Update robots if a new list is provided
+        self.robots = robots
+        self.actions_robots_mapping = actions_robots_mapping
+        self.observations_robots_mapping = observations_robots_mapping
         self.cameras.cameras_ids_to_record = cameras_ids_to_record  # type: ignore
         self.freq = freq  # Store for record_loop
         self.episode_format = episode_format
@@ -125,7 +136,7 @@ class Recorder:
             self.episode = await JsonEpisode.start_new(
                 base_recording_folder=self.episode_recording_folder,
                 dataset_name=dataset_name,
-                robots=self.robots,
+                robots=robots,
                 # Any other necessary params for JsonEpisode metadata
             )
             # if self._use_push_to_hub_after_save:
@@ -139,17 +150,19 @@ class Recorder:
             # LeRobotDataset constructor will ensure directories like meta, data, videos exist.
             lerobot_dataset_manager = LeRobotDataset(path=dataset_full_path)
 
-            # if self._use_push_to_hub_after_save:
-            #     self._current_dataset_full_path_for_push = lerobot_dataset_manager.folder_full_path
-
+            robots_to_initialize = [
+                robots[i] for i in observations_robots_mapping.keys()
+            ]
             self.episode = await LeRobotEpisode.start_new(
                 dataset_manager=lerobot_dataset_manager,  # Pass the dataset manager
-                robots=self.robots,
+                robots=robots_to_initialize,
                 codec=codec,
                 freq=freq,
                 target_size=target_size,
                 instruction=instruction,
                 all_camera_key_names=self.cameras.get_all_camera_key_names(),
+                add_metadata=add_metadata,
+                save_cartesian=save_cartesian,
             )
         else:
             logger.error(f"Unknown episode format: {self.episode_format}")
@@ -168,6 +181,7 @@ class Recorder:
             target_size=target_size,
             language_instruction=instruction
             or config.DEFAULT_TASK_INSTRUCTION,  # Passed to Step
+            save_cartesian=save_cartesian,
         )
         logger.success(
             f"Recording started for {self.episode_format} dataset '{dataset_name}'. Episode index: {self.episode.episode_index if self.episode else 'N/A'}"
@@ -246,9 +260,9 @@ class Recorder:
 
         return None
 
-    def push_to_hub(self, dataset_path: str, branch_path: str | None = None) -> None:
+    def push_to_hub(self, dataset_path: str, branch_path: Optional[str] = None) -> None:
         logger.info(
-            f"Attempting to push dataset from {dataset_path} to Hugging Face Hub. Branch: {branch_path or 'main'}"
+            f"Attempting to push dataset from {dataset_path} to Hugging Face Hub. Will push to 'main', and create branches 'v2.1' and '{branch_path}'if specified."
         )
         try:
             # Dataset class needs to be robust enough to be initialized with the full path
@@ -274,6 +288,9 @@ class Recorder:
         self,
         target_size: tuple[int, int],
         language_instruction: str,  # This is the initial instruction
+        save_cartesian: Optional[
+            bool
+        ] = False,  # Saves cartesian positions if True (only for robots with simulators)
     ) -> None:
         if not self.episode:
             logger.error(
@@ -296,7 +313,7 @@ class Recorder:
 
             # --- Optimized Image Gathering with Parallel Processing ---
             main_frames, secondary_frames = await self._gather_frames_parallel(
-                target_size
+                target_size=target_size,
             )
 
             if main_frames and len(main_frames) > 0:
@@ -308,9 +325,13 @@ class Recorder:
 
             # --- Optimized Robot Observation with Parallel Processing ---
             (
-                final_state,
-                final_joints_position,
-            ) = await self._gather_robot_observations_parallel()
+                final_observation_state,
+                final_observation_joints_position,
+                final_action_state,
+                final_action_joints_position,
+            ) = await self._gather_robot_observations_parallel(
+                save_cartesian=save_cartesian
+            )
 
             current_time_in_episode = loop_iteration_start_time - self.start_ts
 
@@ -326,9 +347,9 @@ class Recorder:
             observation = Observation(
                 main_image=main_frame,
                 secondary_images=secondary_frames,
-                state=final_state,  # Robot's end-effector state(s)
+                state=final_observation_state,  # Robot's end-effector state(s)
                 language_instruction=current_instruction,
-                joints_position=final_joints_position,  # Actual joint positions
+                joints_position=final_observation_joints_position,  # Actual joint positions
                 timestamp=current_time_in_episode,
             )
 
@@ -338,7 +359,8 @@ class Recorder:
             # update_previous_step handles this.
             step = Step(
                 observation=observation,
-                action=None,  # Will be filled by update_previous_step for the *previous* step
+                action=final_action_joints_position,  # Will be filled by update_previous_step for the *previous* step
+                action_cartesian=final_action_state,
                 metadata={"created_at": loop_iteration_start_time},
             )
 
@@ -356,7 +378,9 @@ class Recorder:
                 )
 
             # Order: update previous, then add current.
-            if self.episode.steps:  # If there's a previous step
+            if (
+                self.episode.steps and final_action_joints_position is None
+            ):  # If there's a previous step
                 # The 'action' of the previous step is the 'joints_position' of the current observation
                 self.episode.update_previous_step(step)
 
@@ -370,7 +394,7 @@ class Recorder:
             # Log performance metrics every 100 steps
             if step_count % 100 == 0:
                 logger.debug(
-                    f"Step {step_count}: Processing time: {elapsed_this_iteration:.3f}s, Target: {1/self.freq:.3f}s"
+                    f"Step {step_count}: Processing time: {elapsed_this_iteration:.3f}s, Target: {1 / self.freq:.3f}s"
                 )
 
             await asyncio.sleep(time_to_wait)
@@ -385,7 +409,7 @@ class Recorder:
 
     async def _gather_frames_parallel(
         self, target_size: tuple[int, int]
-    ) -> tuple[List[np.ndarray], List[np.ndarray]]:
+    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
         """
         Simple parallel frame capture - each camera runs independently.
         Returns (main_frames, secondary_frames).
@@ -442,7 +466,7 @@ class Recorder:
         return main_frames, secondary_frames
 
     def _capture_single_camera(
-        self, camera, target_size: tuple[int, int]
+        self, camera: BaseCamera, target_size: tuple[int, int]
     ) -> Optional[np.ndarray]:
         """
         Capture frame from a single camera.
@@ -457,73 +481,140 @@ class Recorder:
             return None
 
     async def _gather_robot_observations_parallel(
-        self,
-    ) -> tuple[np.ndarray, np.ndarray]:
+        self, save_cartesian: Optional[bool] = False
+    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
         """
         Simple parallel robot observation gathering - each robot runs independently.
         Returns (final_state, final_joints_position).
         """
-        if not self.robots:
-            return np.array([]), np.array([])
+        if (
+            not self.robots
+            or not self.observations_robots_mapping
+            or len(self.observations_robots_mapping) == 0
+        ):
+            return np.array([]), np.array([]), np.array([]), np.array([])
 
         loop = asyncio.get_event_loop()
 
         # Submit individual robot observation tasks
-        robot_futures = []
-        for i, robot_instance in enumerate(self.robots):
-            future = loop.run_in_executor(
-                self._robot_thread_pool,
-                self._get_single_robot_observation,
-                robot_instance,
-                i,
+        robot_observations_futures = []
+        robot_actions_future = []
+        for idx, robot in enumerate(self.robots):
+            assert isinstance(robot, BaseRobot), (
+                "Robot must be an instance of BaseRobot."
             )
-            robot_futures.append(future)
+            if idx in self.observations_robots_mapping:
+                source = self.observations_robots_mapping[idx]
+                future_observations_real = loop.run_in_executor(
+                    self._robot_thread_pool,
+                    self._get_single_robot_observation,
+                    robot,
+                    idx,
+                    source,
+                    save_cartesian,
+                )
+                robot_observations_futures.append(future_observations_real)
+            # A robot can be both in action and observation mappings
+            if idx in self.actions_robots_mapping:
+                source = self.actions_robots_mapping[idx]
+                future_actions_real = loop.run_in_executor(
+                    self._robot_thread_pool,
+                    self._get_single_robot_observation,
+                    robot,
+                    idx,
+                    source,
+                    save_cartesian,
+                )
+                robot_actions_future.append(future_actions_real)
 
         # Wait for all robot observations
-        all_robot_states = []
-        all_robot_joints_positions = []
+        all_robots_observation_states = []
+        all_robots_observation_joints_positions = []
+        all_robots_actions_states = []
+        all_robots_actions_joints_positions = []
 
-        for future in robot_futures:
+        for future in robot_observations_futures:
             try:
                 robot_state, robot_joints = await future
                 if robot_state is not None and robot_joints is not None:
-                    all_robot_states.append(robot_state)
-                    all_robot_joints_positions.append(robot_joints)
+                    all_robots_observation_states.append(robot_state)
+                    all_robots_observation_joints_positions.append(robot_joints)
             except Exception as e:
                 logger.warning(f"Failed to get robot observation: {e}")
+        for future in robot_actions_future:
+            try:
+                robot_state, robot_joints = await future
+                if robot_state is not None and robot_joints is not None:
+                    all_robots_actions_states.append(robot_state)
+                    all_robots_actions_joints_positions.append(robot_joints)
+            except Exception as e:
+                logger.warning(f"Failed to get action robot observation: {e}")
 
         # Concatenate if multiple robots, otherwise use the first robot's data
-        final_state = (
-            np.concatenate(all_robot_states)
-            if len(all_robot_states) > 1
-            else (all_robot_states[0] if all_robot_states else np.array([]))
-        )
-        final_joints_position = (
-            np.concatenate(all_robot_joints_positions)
-            if len(all_robot_joints_positions) > 1
+        final_observation_state = (
+            np.concatenate(all_robots_observation_states)
+            if len(all_robots_observation_states) > 1
             else (
-                all_robot_joints_positions[0]
-                if all_robot_joints_positions
+                all_robots_observation_states[0]
+                if all_robots_observation_states
+                else np.array([])
+            )
+        )
+        final_observation_joints_position = (
+            np.concatenate(all_robots_observation_joints_positions)
+            if len(all_robots_observation_joints_positions) > 1
+            else (
+                all_robots_observation_joints_positions[0]
+                if all_robots_observation_joints_positions
+                else np.array([])
+            )
+        )
+        final_action_state = (
+            np.concatenate(all_robots_actions_states)
+            if len(all_robots_actions_states) > 1
+            else (
+                all_robots_actions_states[0]
+                if all_robots_actions_states
+                else np.array([])
+            )
+        )
+        final_action_joints_position = (
+            np.concatenate(all_robots_actions_joints_positions)
+            if len(all_robots_actions_joints_positions) > 1
+            else (
+                all_robots_actions_joints_positions[0]
+                if all_robots_actions_joints_positions
                 else np.array([])
             )
         )
 
-        return final_state, final_joints_position
+        return (
+            final_observation_state,
+            final_observation_joints_position,
+            final_action_state,
+            final_action_joints_position,
+        )
 
     def _get_single_robot_observation(
-        self, robot, robot_idx: int
-    ) -> tuple[Optional[np.ndarray], Optional[np.ndarray]]:
+        self,
+        robot: BaseRobot,
+        robot_idx: int,
+        source: Literal["sim", "robot"],
+        do_forward: Optional[bool] = False,
+    ) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
         """
         Get observation from a single robot.
         This runs in the thread pool.
         """
         try:
-            return robot.get_observation()
+            return robot.get_observation(
+                source=source, do_forward=do_forward if do_forward else False
+            )
         except Exception as e:
             logger.warning(f"Exception getting observation from robot {robot_idx}: {e}")
             return None, None
 
-    def __del__(self):
+    def __del__(self) -> None:
         """Cleanup thread pools on deletion."""
         if hasattr(self, "_image_thread_pool") and self._image_thread_pool:
             self._image_thread_pool.shutdown(wait=True)
