@@ -1,12 +1,13 @@
 import asyncio
 import json
 import time
+from typing import Optional
 
 from fastapi import HTTPException
 from loguru import logger
 from supabase import AsyncClient, acreate_client
-from supabase_auth.types import Session as SupabaseSession
 from supabase_auth.errors import AuthRetryableError
+from supabase_auth.types import Session as SupabaseSession
 
 from phosphobot.models import Session
 from phosphobot.utils import get_home_app_path, get_tokens
@@ -43,7 +44,7 @@ def save_session(session: Session) -> None:
         json.dump(session.model_dump(), f)
 
 
-def load_session() -> Session | None:
+def load_session() -> Optional[Session]:
     """
     Load the session from a file.
     """
@@ -75,8 +76,11 @@ async def get_client() -> AsyncClient:
     session = load_session()
 
     async def set_session_with_retry(
-        access_token, refresh_token, max_retries=3, delay=2
-    ):
+        access_token: str,
+        refresh_token: str,
+        max_retries: int = 3,
+        delay: float = 2,
+    ) -> bool:
         current_delay = delay
         for attempt in range(max_retries):
             try:
@@ -94,6 +98,7 @@ async def get_client() -> AsyncClient:
                 else:
                     logger.error(f"Failed after {max_retries} attempts: {e}")
                     return False
+        return False
 
     if session:
         if (
